#include "LocalExpansionCoefficients.hpp"
#include "MultipoleExpansionCoefficients.hpp"

LocalExpansionCoefficients::LocalExpansionCoefficients(Vec3f r,Vec3f nd,float radiosity,Vec3f _center,int _p)
                           :ExpansionCoefficients(_center,_p){
	r=r-center;
	Vec3f rs=sph(r);
	SourceMatrices sm(r,nd);
	fillYnmCache(rs[1],rs[2],p);
	for(int n=0;n<=p;n++){
		for(int j=0;j<=n/2;j++){
			for(int m=-n+2*j;m<=n-2*j;m++){
			  csms[n][j][m]=e[n][j]*radiosity*pow(rs[0],-n-4)*conj(Ynm[n-2*j][m])*sm;
			}
		}
	}
}

LocalExpansionCoefficients LocalExpansionCoefficients::translateTo(Vec3f newCenter){
	LocalExpansionCoefficients l(newCenter,p);
	Vec3f x1x0=center-newCenter;
	Vec3f x1x0s=sph(x1x0);
	TranslationMatrices tm(-x1x0);
	fillYnmCache(x1x0s[1],x1x0s[2],p);
	for(int kd=0;kd<=p;kd++){
		for(int sd=0;sd<=kd/2;sd++){
			for(int m1d=-kd+2*sd;m1d<=kd-2*sd;m1d++){
				for(int k=kd;k<=p;k++){
					for(int s=0;s<=k/2;s++){
						for(int m1=-k+2*s;m1<=k-2*s;m1++){
							int jmin=max(s-kd+sd,0);
							int jmax=min(s-sd,k-kd+sd-s);
							for(int j=jmin;j<=jmax;j++){
								if ((abs(m1d)>(kd-2*sd))||(abs(-m1+m1d)>(k-kd-2*j))) break;
								l.csms[kd][sd][m1d]+=(m1%2==0?1:-1)*
									(k%2==0?1:-1)*
									(e[k-kd][j]/e[k][s])*
									J(kd-2*sd,m1d,k-2*s,sd,-k+kd-4,-m1+m1d,k-kd-2*j)*
									(tm*((pow(x1x0s[0],k-kd)*conj(Ynm[k-kd-2*j][-m1+m1d]))*csms[k][s][m1]));
							}
						}
					}
				}
			}
		}
	}
	
	return l;
}

float LocalExpansionCoefficients::evaluateAtPoint(Vec3f x,Vec3f n,float rho){
	Vec3f x0x=x-center;
	Vec3f x0xs=sph(x0x);
	ReceiverMatrices rm(x0x,n);
	CplxReceiverMatrices crm=Cplxf(1,0)*rm;
	Cplxf ans(0,0);
	fillYnmCache(x0xs[1],x0xs[2],p);
	for(int k=0;k<=p;k++){
		for(int s=0;s<=k/2;s++){
			for(int m1=-k+2*s;m1<=k-2*s;m1++){
				ans+=(pow(x0xs[0],k)*Ynm[k-2*s][m1])*(crm*csms[k][s][m1]);
			}
		}
	}
	return rho*real(ans);
}

void LocalExpansionCoefficients::addPatchPointLocal(PatchPoint* pp,int channel){
	Vec3f r=pp->coords-center;
	Vec3f rs=sph(r);
	SourceMatrices sm(r,pp->normal);
	fillYnmCache(rs[1],rs[2],p);
	for(int n=0;n<=p;n++){
		for(int j=0;j<=n/2;j++){
			for(int m=-n+2*j;m<=n-2*j;m++){
			  csms[n][j][m]+=e[n][j]*pp->weight*pp->radVal[channel]*pow(rs[0],-n-4)*conj(Ynm[n-2*j][m])*sm;
			}
		}
	}
}
 

void LocalExpansionCoefficients::addTranslatedLocal(LocalExpansionCoefficients* lecs){
	Vec3f x1x0=lecs->center-center;
	Vec3f x1x0s=sph(x1x0);
	TranslationMatrices tm(-x1x0);
	fillYnmCache(x1x0s[1],x1x0s[2],p);
	for(int kd=0;kd<=p;kd++){
		for(int sd=0;sd<=kd/2;sd++){
			for(int m1d=-kd+2*sd;m1d<=kd-2*sd;m1d++){
				for(int k=kd;k<=p;k++){
					for(int s=0;s<=k/2;s++){
						for(int m1=-k+2*s;m1<=k-2*s;m1++){
							int jmin=max(s-kd+sd,0);
							int jmax=min(s-sd,k-kd+sd-s);
							for(int j=jmin;j<=jmax;j++){
								if ((abs(m1d)>(kd-2*sd))||(abs(-m1+m1d)>(k-kd-2*j))) break;
								 csms[kd][sd][m1d]+=(m1%2==0?1:-1)*
									(k%2==0?1:-1)*
									(e[k-kd][j]/e[k][s])*
									J(kd-2*sd,m1d,k-2*s,sd,-k+kd-4,-m1+m1d,k-kd-2*j)*
									(tm*((pow(x1x0s[0],k-kd)*conj(Ynm[k-kd-2*j][-m1+m1d]))*lecs->csms[k][s][m1]));
							}
						}
					}
				}
			}
		}
	}
}

void LocalExpansionCoefficients::addTranslatedMultipole(MultipoleExpansionCoefficients* mecs){
	Vec3f ox0=center-mecs->center;
	Vec3f ox0s=sph(ox0);
	TranslationMatrices tm(ox0);
	fillYnmCache(ox0s[1],ox0s[2],2*p);
	for(int k=0;k<=p;k++){
		for(int s=0;s<=k/2;s++){
			for(int m1=-k+2*s;m1<=k-2*s;m1++){
				for(int n=0;n<=p;n++){
					for(int j=0;j<=n/2;j++){
						for(int m=-n+2*j;m<=n-2*j;m++){
							int lambdamin=max(abs(n-2*j-k+2*s),abs(m-m1));
							int lmax=n-2*j+k-2*s;
							int lmin=((n-2*j-k+2*s-lambdamin)%2==0?(lambdamin):(lambdamin+1));
							for(int l2=lmin;l2<=lmax;l2+=2){
								csms[k][s][m1]+=tm*(J(k-2*s,m1,l2,s,-n-4,m,n-2*j)*e[n][j]*pow(ox0s[0],-n-4-k)*Ynm[l2][m-m1]*mecs->csms[n][j][m]);
							}
						}
					}
				}
			}
		}
	}
}
