#include "MathFuncs.hpp"

int PMAX=0;
float **e;
float *******jstore;

void initMath(int p){
	PMAX=p;
	initSphHrm();
	ecache(p);
	jcache(p);
	initIntegration();
}

Vec3f sph(const Vec3f& v){
  Vec3f retVal(0,0,0);
  retVal[0]=len(v);
  retVal[2]=atan2(v[1],v[0]);
  retVal[1]=acos(v[2]/len(v));
  return retVal;
}

float fact(float n){
  if (n<=0)
    return 1;
  return n*fact(n-1);
}

void ecache(int p){
  e=new float*[p+1];
  for(int i=0;i<=p;i++)
    e[i]=new float[p+1];
  for(int n=0;n<=p;n++)
    for(int j=0;j<=n/2;j++){
      e[n][j]=(4*M_PI/(2*(n-2*j)+1))*(2*n-4*j+1)*fact(n-j+1)*fact(j+0.5)/(fact(n-j+0.5)*fact(j));
    }
}

void jcache(int p){
#ifdef JCACHE
	jstore=new float******[p+1];
  for(int l1=0;l1<=p;l1++){
    jstore[l1]=new float*****[2*p+1];
    for(int m1=-p;m1<=p;m1++){
      jstore[l1][m1+p]=new float****[2*p+1];
      for(int l2=0;l2<=2*p;l2++){
	jstore[l1][m1+p][l2]=new float***[p/2+1];
	for(int s=0;s<=p/2;s++){
	  jstore[l1][m1+p][l2][s]=new float**[p+4+1];
	  for(int N=-p-4;N<=0;N++){
	    jstore[l1][m1+p][l2][s][N+p+4]=new float*[2*p+1];
	    for(int M=-p;M<=p;M++){
	      jstore[l1][m1+p][l2][s][N+p+4][M+p]=new float[p];
	      for(int L=0;L<=p;L++){
		jstore[l1][m1+p][l2][s][N+p+4][M+p][L]=Jcalc(l1,m1,l2,s,N,M,L);
		//		cout<<jstore[l1][m1+p][l2][s][N+p+4][M+p][L]<<endl;
	      }
	    }
	  }
	}
      }
    }
  }
#endif
}

float poc(float a,float s){
  float ans=1;
  for(int i=1;i<=s;i++)
    ans*=(a+i-1);
  return ans;
}
 
#ifndef JCACHE
float J(int l1,int m1,int l2,int s,int N,int M,int L){
  int n=N-L;
  int lambda=(-L+l1+l2)/2;
  int lambda1=(L-l1+l2)/2;
  float wigner=2*M_PI*((M%2)==0?1:-1)*sqrtf((2*L+1)*(2*l1+1)*(2*l2+1)/(4*M_PI))*
    gsl_sf_coupling_3j((int)(2*L),(int)(2*l1),(int)(2*l2),(int)(-2*M),(int)(2*m1),(int)(2*(M-m1)))*
    gsl_sf_coupling_3j((int)(2*L),(int)(2*l1),(int)(2*l2),0,0,0);
  /*l1*/float numerator=((lambda)%2==0?1:-1)*wigner*
	  poc(-n/2.0f,lambda)*poc((n+3)/2.0f,L)*poc(lambda-n/2.0f,s)*poc(-0.5-n/2.0f-lambda1,s);
  float denominator=poc(0.5,l1+1)*poc((n+3)/2.0f,lambda1)*poc(l1+1.5,s)*fact(s);
  return numerator/denominator;
}
float Jcalc(int l1,int m1,int l2,int s,int N,int M,int L){
 return 0;
} 
#else
float Jcalc(int l1,int m1,int l2,int s,int N,int M,int L){
  int n=N-L;
  int lambda=(-L+l1+l2)/2;
  int lambda1=(L-l1+l2)/2;
  float wigner=2*M_PI*((M%2)==0?1:-1)*sqrtf((2*L+1)*(2*l1+1)*(2*l2+1)/(4*M_PI))*
    gsl_sf_coupling_3j((int)(2*L),(int)(2*l1),(int)(2*l2),(int)(-2*M),(int)(2*m1),(int)(2*(M-m1)))*
    gsl_sf_coupling_3j((int)(2*L),(int)(2*l1),(int)(2*l2),0,0,0);
  /*l1*/float numerator=((lambda)%2==0?1:-1)*wigner*
	  poc(-n/2.0f,lambda)*poc((n+3)/2.0f,L)*poc(lambda-n/2.0f,s)*poc(-0.5-n/2.0f-lambda1,s);
  float denominator=poc(0.5,l1+1)*poc((n+3)/2.0f,lambda1)*poc(l1+1.5,s)*fact(s);
  return numerator/denominator;
}
float J(int l1,int m1,int l2,int s,int N,int M,int L){
	 return jstore[l1][m1+PMAX][l2][s][N+PMAX+4][M+PMAX][L];
}
#endif
