#ifndef __matrixsets__
#define __matrixsets__

#include "src/Radical.hpp"
#include "Vec3ft.hpp"
#include "vl/Mat3.h"

class SourceMatrices {
 public:
  Mat3f m1;
  Vec3f m2;
  Vec3ft m3;
  float m4;
  
  SourceMatrices(){}
  SourceMatrices(Vec3f r,Vec3f n){
    Vec3ft rt=trans(r);
    m1=n*rt;
    m2=n;
    m3=rt*n*rt;
    m4=rt*n;
  }

  void operator+=(SourceMatrices sm){           // A+=B
    m1+=sm.m1;
    m2+=sm.m2;
    m3+=sm.m3;
    m4+=sm.m4;
  }
  
  SourceMatrices operator+(SourceMatrices sm){       // C = A+B
    sm.m1+=m1;
    sm.m2+=m2;
    sm.m3+=m3;
    sm.m4+=m4;
    return sm;
  }
  SourceMatrices operator-(SourceMatrices sm){       //C = A-B
    sm.m1=m1-sm.m1;
    sm.m2=m2-sm.m2;
    sm.m3=m3-sm.m3;
    sm.m4=m4-sm.m4;
    return sm;
  }
};

class ReceiverMatrices {
 public:
  Vec3ft m1;
  Vec3f m2;
  Vec3ft m3;
  float m4;
  Mat3f m5;
  Vec3ft m6;

  ReceiverMatrices(){}
  ReceiverMatrices(Vec3f r,Vec3f n){
    Vec3ft rt=trans(r);
    m1=rt;
    m2=n;
    m3=rt*n*rt;
    m4=rt*n;
    m5=n*rt;
    m6=n;
  }
  void operator+=(ReceiverMatrices rm){
    m1+=rm.m1;
    m2+=rm.m2;
    m3+=rm.m3;
    m4+=rm.m4;
    m5+=rm.m5;
  }

  ReceiverMatrices operator+(ReceiverMatrices rm){
    rm.m1+=m1;
    rm.m2+=m2;
    rm.m3+=m3;
    rm.m4+=m4;
    rm.m5+=m5;
    return rm;
  }
  ReceiverMatrices operator-(ReceiverMatrices rm){
    rm.m1=m1-rm.m1;
    rm.m2=m2-rm.m2;
    rm.m3=m3-rm.m3;
    rm.m4=m4-rm.m4;
    rm.m5=m5-rm.m5;
    return rm;
  }
};

class TranslationMatrices {
 public:
  Vec3ft m1;
  Mat3f m2;

  TranslationMatrices(){}
  TranslationMatrices(Vec3f r){
    Vec3ft rt=trans(r);
    m1=rt;
    m2=r*rt;
  }
};

typedef Complex<SourceMatrices>   CplxSourceMatrices;
typedef Complex<ReceiverMatrices> CplxReceiverMatrices;




inline float operator*(ReceiverMatrices rm,SourceMatrices sm){
  return rm.m1*(sm.m1*rm.m6)-rm.m3*sm.m2-sm.m3*rm.m2+sm.m4*rm.m4;
}

inline ReceiverMatrices operator*(float p,ReceiverMatrices rm){
  ReceiverMatrices rm2;
  rm2.m1=rm.m1*p;
  rm2.m6=rm.m6;
  rm2.m2=rm.m2*p;
  rm2.m3=rm.m3*p;
  rm2.m4=rm.m4*p;
  rm2.m5=rm.m5*p;
  return rm2;
}

inline SourceMatrices operator*(float p, SourceMatrices sm){
  SourceMatrices sm2;
  sm2.m1=sm.m1*p;
  sm2.m2=sm.m2*p;
  sm2.m3=sm.m3*p;
  sm2.m4=sm.m4*p;
  return sm2;
}

inline SourceMatrices operator*(TranslationMatrices tm, SourceMatrices sm){
  SourceMatrices sm2;
  sm2.m1=sm.m1-sm.m2*tm.m1;
  sm2.m2=sm.m2;
  sm2.m3=sm.m3-sm.m4*tm.m1-tm.m1*sm.m1+trans(sm.m2)*tm.m2;
  sm2.m4=sm.m4-tm.m1*sm.m2;
  return sm2;
}

inline ReceiverMatrices operator*(ReceiverMatrices rm,TranslationMatrices tm){
  ReceiverMatrices rm2;
  rm2.m1=rm.m1+tm.m1;
  rm2.m2=rm.m2;
  rm2.m6=rm.m6;
  rm2.m3=rm.m3+rm.m4*tm.m1+tm.m1*rm.m5+trans(rm.m2)*tm.m2;
  rm2.m4=rm.m4+tm.m1*rm.m2;
  rm2.m5=rm.m5+rm.m2*tm.m1;;
  return rm2;
}

inline CplxSourceMatrices operator*(Cplxf mult,SourceMatrices m){
  return CplxSourceMatrices(real(mult)*m,imag(mult)*m);
}

inline CplxReceiverMatrices operator*(Cplxf mult,ReceiverMatrices m){
  return CplxReceiverMatrices(real(mult)*m,imag(mult)*m);
}

inline Cplxf operator*(CplxReceiverMatrices rm,CplxSourceMatrices sm){
  return Cplxf(real(rm)*real(sm)-imag(rm)*imag(sm),real(rm)*imag(sm)+imag(rm)*real(sm));
}

inline CplxSourceMatrices operator*(float p, CplxSourceMatrices sm){
  return CplxSourceMatrices(p*real(sm),p*imag(sm));
}

inline CplxReceiverMatrices operator*(float p, CplxReceiverMatrices rm){
  return CplxReceiverMatrices(p*real(rm),p*imag(rm));
}

inline CplxSourceMatrices operator*(Cplxf z, CplxSourceMatrices sm){
  return CplxSourceMatrices(real(z)*real(sm)-imag(z)*imag(sm),real(z)*imag(sm)+imag(z)*real(sm));
}

inline CplxReceiverMatrices operator*(Cplxf z, CplxReceiverMatrices rm){
  return CplxReceiverMatrices(real(z)*real(rm)-imag(z)*imag(rm),real(z)*imag(rm)+imag(z)*real(rm));
}

inline CplxSourceMatrices operator*(TranslationMatrices tm, CplxSourceMatrices sm){
  return CplxSourceMatrices(tm*real(sm),tm*imag(sm));
}

inline CplxReceiverMatrices operator*(CplxReceiverMatrices rm,TranslationMatrices tm){
  return CplxReceiverMatrices(real(rm)*tm,imag(rm)*tm);
}

/*

inline ostream &operator<<(ostream &s, const SourceMatrices &m){
  return (s<<"["<<endl<<m.m1<<endl<<m.m2<<endl<<m.m3<<endl<<m.m4<<endl<<"]");
}

inline ostream &operator<<(ostream &s, const ReceiverMatrices &m){
  return (s<<"["<<endl<<m.m1<<endl<<m.m2<<endl<<m.m3<<endl<<m.m4<<endl<<m.m5<<endl<<"]");
}

inline ostream &operator<<(ostream &s, const TranslationMatrices &m){
  return (s<<"["<<endl<<m.m1<<endl<<m.m2<<"]");
}

inline ostream &operator<<(ostream &s, const CplxReceiverMatrices &m){
  return (s<<"["<<real(m)<<endl<<","<<endl<<imag(m)<<"]");
}

inline ostream &operator<<(ostream &s, const CplxSourceMatrices &m){
  return (s<<"["<<real(m)<<endl<<","<<endl<<imag(m)<<"]");
}
*/

#endif
