#include "Cells.hpp"
//#include<leaves.hpp>
//#include <GL/glut.h>
//#include <vl/VLgl.h>
#include <stdlib.h>
#include<math.h>
#include "PatchPoint.hpp"
#include <fstream>

#define epsilon .000001
#define DIFF_ANGLE 20
#define max 1000
#define CLOSEST_COUNT 3
#define RADIUS 0.5
#define CELL_LENGTH 2.0
//#define DELTA CELL_LENGTH

#define POINT_POINT_THRESHOLD .22   // Random :-)
#define POINT_CELL_THRESHOLD .4
#define CELL_CELL_THRESHOLD .4
#define PI vl_pi

vector<Cell *> leafcell;
/*Vec3f axesDisplace[]={
  Vec3f(-0.5,-0.5,-0.5),
  Vec3f(-0.5,-0.5,0.5),
  Vec3f(-0.5,0.5,-0.5),
  Vec3f(0.5,-0.5,-0.5),
  Vec3f(0.5,0.5,0.5),
  Vec3f(0.5,-0.5,0.5),
  Vec3f(-0.5,0.5,0.5),
  Vec3f(0.5,0.5,-0.5)
};*/
Vec3f axesDisplace[]={       //for drawing visible cells
  Vec3f(-0.5,-0.5,-0.5),
  Vec3f(-0.5,0.5,-0.5),
  Vec3f(-0.5,0.5,0.5),
  Vec3f(-0.5,-0.5,0.5),
  Vec3f(0.5,-0.5,0.5),
  Vec3f(0.5,0.5,0.5),
  Vec3f(0.5,0.5,-0.5),
  Vec3f(0.5,-0.5,-0.5)
};
Vec3f planes[]={
  Vec3f(-0.5,-0.5,0.5),
  Vec3f(-0.5,-0.5,-0.5),
  Vec3f(-0.5,0.5,0.5),     //plane 1 -- left
  Vec3f(-0.5,-0.5,0.5),
  Vec3f(0.5,-0.5,0.5),
  Vec3f(0.5,0.5,0.5),      //plane 2 -- front
  Vec3f(0.5,-0.5,0.5),
  Vec3f(0.5,0.5,0.5),
  Vec3f(0.5,-0.5,-0.5),    //plane 3 -- right
  Vec3f(-0.5,0.5,-0.5),
  Vec3f(-0.5,-0.5,-0.5), 
  Vec3f(0.5,-0.5,-0.5),	   //plane 1 -- back
  Vec3f(0.5,0.5,-0.5),
  Vec3f(0.5,0.5,0.5),
  Vec3f(-0.5,0.5,-0.5),	   //plane 2 -- top
  Vec3f(0.5,-0.5,0.5),
  Vec3f(-0.5,-0.5,-0.5),
  Vec3f(0.5,-0.5,-0.5)	  //plane 2 -- botom
};

long maxDepth = -1;

double findT(Vec3f,Vec3f,Vec3f,Vec3f,Vec3f);
bool check_condition(Vec3f, Vec3f, Vec3f, Cell*);
double findN(float,float,float,float,float,float,float,float,float,float,float,float,float,float,float);
double findD(float,float,float,float,float,float,float,float,float,float,float,float,float,float,float);
int find_max(vector<float>&); 
float find_distance(Vec3f, Vec3f, Vec3f);
float visibilityLookUp(float);

extern Cell* root;
extern vector<Vec3f> cell_centers;
extern float delta;

/*****************************************************************************/
/*****************************************************************************/

float visibilityLookUp(float dist) // THE VISIBILITY VALUE DEPENDS ON THE THRESHOLD WHICH IN TURN DEPENDS ON CLOSEST_COUNT
{
  if(dist <=1e-6)
	return (0);
  else if(dist > 0.0 && dist <= (0.1*delta))
	return (0.25);
  else if(dist > (0.1*delta) && dist <= (0.2*delta))
	return (0.55);
  else if(dist > (0.2*delta) && dist <= (0.4*delta))
	return (0.7);
  else if(dist > (0.4*delta) && dist <= (0.8*delta))
	return (0.8);
  else if(dist > (0.8*delta) && dist < delta)
	return (0.99);
  else
       return(1);
}

/*****************************************************************************/
/*****************************************************************************/

double findN(float x1,float y1,float z1,float x2,float y2,float z2,float x3,float y3,float z3,float x4,float y4,float z4,float x5,float y5,float z5)
{
  return(x2*(y3*z4 - z3*y4) - x3*(y2*z4 - z2*y4) + x4*(y2*z3 - z2*y3) - x1*(y3*z4 - z3*y4) + x3*(y1*z4 - z1*y4) - x4*(y1*z3 - z1*y3) + x1*(y2*z4 - z2*y4) - x2*(y1*z4 - z1*y4) + x4*(y1*z2 - z1*y2) - x1*(y2*z3 - z2*y3) + x2*(y1*z3 - y3*z1) + x3*(y1*z2 - z1*y2));

}

/*****************************************************************************/
/*****************************************************************************/

double findD(float x1,float y1,float z1,float x2,float y2,float z2,float x3,float y3,float z3,float x4,float y4,float z4,float x5,float y5,float z5)
{
  return(x2*(y3*(z5-z4) - z3*(y5-y4)) - x3*(y2*(z5-z4) - z2*(y5-y4)) + (x5-x4)*(y2*z3 - z2*y3) - x1*(y3*(z5-z4) - z3*(y5-y4)) + x3*(y1*(z5-z4) - z1*(y5-y4)) - (x5-x4)*(y1*z3 - z1*y3) + x1*(y2*(z5-z4) - z2*(y5-y4)) - x2*(y1*(z5-z4) - z1*(y5-y4)) + (x5-x4)*(y1*z2 - z1*y2));

}

/*****************************************************************************/
/*****************************************************************************/

double findT(Vec3f X1,Vec3f X2,Vec3f X3,Vec3f c1,Vec3f c2)
{
  double numerator, denominator, t;
  float x1 = X1[0];
  float y1 = X1[1];
  float z1 = X1[2];
  float x2 = X2[0];
  float y2 = X2[1];
  float z2 = X2[2];
  float x3 = X3[0];
  float y3 = X3[1];
  float z3 = X3[2];
  float x4 = c1[0];
  float y4 = c1[1];
  float z4 = c1[2];
  float x5 = c2[0];
  float y5 = c2[1];
  float z5 = c2[2];
  numerator   = findN(x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,x5,y5,z5);
  denominator = findD(x1,y1,z1,x2,y2,z2,x3,y3,z3,x4,y4,z4,x5,y5,z5);
  //	printf("numerator: %f \t denominator: %f\n",numerator,denominator);
  t = -1*numerator/(denominator + 0.00000001);
  return (t);
}

/*****************************************************************************/
/*****************************************************************************/

bool check_condition(Vec3f c1, Vec3f c2, Cell* temp)
{
  bool y = temp->contains(c1);
  if(y == 1){//printf("In condition 1\n");
    return (1);}
  y = temp->contains(c2);
  if(y == 1){//printf("In condition 2\n");
    return (1);}
  double t; 
  float sideLength = temp->side;
  Vec3f X1, X2, X3, cr = temp->center;
  for(int j=0;j<16;j+=3)
    {
      X1 = cr + (planes[j]*sideLength);
      X2 = cr + (planes[j+1]*sideLength);
      X3 = cr + (planes[j+2]*sideLength);
      t  = findT(X1, X2, X3, c1, c2);
      Vec3f intersect_pt;
      intersect_pt[0] = c1[0] + (c2[0] - c1[0])*t;
      intersect_pt[1] = c1[1] + (c2[1] - c1[1])*t;
      intersect_pt[2] = c1[2] + (c2[2] - c1[2])*t;
      bool y = temp->contains(intersect_pt);
      if(y == 1){ 
	//				printf("Intersection point is : %f %f %f\n",intersect_pt[0],intersect_pt[1],intersect_pt[2]);
	return (1);
      }
    }
  return(0);
}

/*****************************************************************************/
/*****************************************************************************/

void Cell::check_intersect(Vec3f c1, Vec3f c2, vector<float>& distance) {
  if(depth == 0) {
    for(int i=0;i<children.size();i++) {
      children[i]->check_intersect(c1, c2, distance ); }}
  else {
    bool y = check_condition(c1,c2, this);
    if(y == 1) {
      if(!isLeaf()){
	for(int i=0;i<children.size();i++) {
	  children[i]->check_intersect(c1, c2, distance); }}
      else {
	for(PPListItr point_itr = patchPoints.begin();point_itr != patchPoints.end();point_itr++) {
	  Vec3f px = (*point_itr)->coords;
	  Vec3f pn = (*point_itr)->normal;
	  if(((*point_itr)->coords==c1) || ((*point_itr)->coords==c2)) {
	    continue;
	  }
	  else if(!(((dot(c1-px,pn) > 0) && (dot(c2-px,pn) < 0)) || ((dot(c1-px,pn) < 0) && (dot(c2-px,pn) > 0)))){
         	continue;
	  }
	  float cosTheta = fabs(dot(pn,norm(c1-c2)));
	  if(cosTheta < 1e-6){
		continue;
	  }
	  float dist = find_distance(c1,c2,px)/cosTheta;

#ifdef MYDEBUG
	  cout << "(C1 C2 Point) = (" << c1 << c2 << (*point_itr)->coords << ") == " << dist << endl;
#endif

	  if(distance.size()<CLOSEST_COUNT) {
	    distance.push_back(dist);
	  }
	  else {
	    int index = find_max(distance);
	    if(dist<distance[index]) {
	      distance[index] = dist;
	    }
	  }
	}
      }
    }
  }
}

/*****************************************************************************/
/*****************************************************************************/

float find_distance(Vec3f c1, Vec3f c2, Vec3f tpoint) {
  /*http://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html*/
  /*distance = |(c2-c1)X(c1-tpoint)|/|c2-c1|*/

  return len(cross(c2-c1,c1-tpoint))/len(c2-c1);
}

/*****************************************************************************/
/*****************************************************************************/




int find_max(vector<float>& distance) {
  float val = -HUGE_VAL;
  int index = -1;
  for(int i=0; i<distance.size(); i++) {
    if(distance[i]>val) {
      val = distance[i];
      index = i;
    }
  }
  return index;
}

/*****************************************************************************/
/*****************************************************************************/

int point_point_visibility(PatchPoint *p, PatchPoint* q) {
  Vec3f pc = p->coords;
  Vec3f qc = q->coords;
  
  Vec3f d = pc - qc;
  
  Vec3f pn = p->normal;
  Vec3f qn = q->normal;
//  Vec3f redWall(1.0,0.0,0.0);
  
#ifdef DEBUG
/*  if((p->mark==1) || (q->mark==1)) {
        if(p->mark==1){
		if(q->normal == redWall){
			cout<<"Dot Products are :: "<<fabs(dot(d,qn))<<"  And  "<<fabs(dot(-d,pn))<<endl;
		 }
	}
	if(q->mark==1){
                if(p->normal == redWall){
                        cout<<"Dot Products are :: "<<fabs(dot(d,qn))<<"  And  "<<fabs(dot(-d,pn))<<endl;
                 }
        }

  }*/
#endif

  /*Near Line Points contain all the points which are within
    a delta from the line connecting */
  vector<float> near_line_distance;
  float visi = 1.0;
//  if((fabs(dot(d,qn)) > 1e-6) && (fabs(dot(-d,pn)) > 1e-6)) {
  if(dot(d,qn) > 0.0 && dot(-d,pn) > 0.0){
    
    root->check_intersect(pc,qc,near_line_distance);
    for(int i=0;i<near_line_distance.size();i++) {
      visi *= visibilityLookUp(near_line_distance[i]);  
#ifdef DEBUG
      //cout << "Near Line Distance :: " << val << endl;
#endif
#ifdef NDEBUG
      if((p->mark==1) || (q->mark==1)) {
	if(p->mark==1)
	  cout<<"Normal is :: "<<q->normal<<"  "<<"Co-ordinate is :: "<<q->coords<<endl;
	if(q->mark==1)
	  cout<<"Normal is :: "<<p->normal<<"  "<<"Co-ordinate is :: "<<p->coords<<endl;
	cout << "Distance is " << near_line_distance[i] <<endl;
      }
#endif

    }

#ifdef NDEBUG
	if((p->mark==1) || (q->mark==1))
	cout<<endl<<endl;
#endif
	
	
#ifdef DEBUG
	/*  if(visi>POINT_POINT_THRESHOLD) {
	    if((p->mark==1) || (q->mark==1)) {
	    if(p->mark==1){
	    if(q->normal == redWall){
	    cout<<"Visibility Value is :: "<<visi<<endl;
	    }
	    }
	    if(q->mark==1){
	    if(p->normal == redWall){
	    cout<<"Visibility Value is :: "<<visi<<endl;
	    }
	    }
	    
	    }
	    }
	*/
    if(visi>POINT_POINT_THRESHOLD) {
	    if((p->mark==1) || (q->mark==1)) {
		if(p->mark==1)
		      cout << "Visible " << qc << " :: " << qn<< "  ::  "<< visi << endl;

		else
		       cout << "Visible " << pc << " :: " << pn<< "  ::  "<< visi << endl;
	    }
    }
    else{
      if((p->mark==1) || (q->mark==1)) {
		if(p->mark==1)
                      cout << "InVisible " << qc << " :: " << qn<< "  ::  "<< visi << endl;

                else
                       cout << "InVisible " << pc << " :: " << pn<< "  ::  "<< visi << endl;
            }
    }
      //cout << "Visi :: " << visi << endl;
   

 
#endif
      
  if(visi>POINT_POINT_THRESHOLD) {
	return 1;
    }
  }
  return 0;
}

int point_cell_visibility(PatchPoint *p, Cell* c) {
  int point_cell_count = 0;
  
  for(PPListItr pitr=c->patchPoints.begin();pitr!=c->patchPoints.end();pitr++) {
    point_cell_count += point_point_visibility(p,(*pitr));
  }
  if(((1.0 * point_cell_count)/c->patchPoints.size())>=POINT_CELL_THRESHOLD) {
    return 1;
  }
  return 0;
}

int cell_cell_visibility(Cell* c, Cell* d) {
  int cell_cell_count = 0;
  for(PPListItr pitr=c->patchPoints.begin();pitr!=c->patchPoints.end();pitr++) {
    cell_cell_count += point_cell_visibility((*pitr),d);
  }
  if(((1.0 * cell_cell_count)/c->patchPoints.size())>=CELL_CELL_THRESHOLD) {
    return 1;
  }
  return 0;
}

int visi_leaf(int s1, int e1, int s2, int e2) {
  int total=0;
  for(int i=s1;i<=e1;i++) {
    for(int j=s2;j<=e2;j++) {
      total += cell_cell_visibility(leafcell[i],leafcell[j]);
    }
  }
  
#ifdef DEBUG
  
#endif
  
  if(total==(e1-s1+1)*(e2-s2+1))return(1);
  if(total==0)return(-1);
  else return(0);
}

// int visi_leaf(int s1, int e1, int s2, int e2, float delta, vector<float>& cell_centers)
// {
//   int total=0;
//   //  printf("(e1-s1+1)*(e2-s2+1) =  %d\n",(e1-s1+1)*(e2-s2+1)); 
//   for(int i=s1;i<=e1;i++) {
//     for(int j=s2;j<=e2;j++) {
//       int visible_count = 0;
//       //for(PPListItr p_itr = leafcell[i]->patchPoints.begin();p_itr != leafcell[i]->patchPoints.end();p_itr++) {
// 	      double v_itr = 0.0;
// 	      //Vec3f c1 = (*p_itr)->coords;
// 	      for(PPListItr point_itr = leafcell[j]->patchPoints.begin();point_itr != leafcell[j]->patchPoints.end();point_itr++)
// 		{
// 		  vector<float> temp_dist(CLOSEST);
// 		  for(int in = 0; in < temp_dist.size(); in++) 
// 		    temp_dist[in] = 0.0;
// 		  float sum=0;
// 		  //Vec3f c2 = (*point_itr)->coords;
// 		  //	printf("Point1 : %f %f %f \t Point2: %f %f %f\n",(*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*point_itr)->coords[0], (*point_itr)->coords[1],(*point_itr)->coords[2]);
// 		  //Vec3f d = c1 - c2;
// 		  float mini = delta+0.000001, total_visible = 1; //Total_visible ka value to be deleted
// 		  //if(((dot(d,(*point_itr)->normal)) < 0) || ((dot(-(d), (*p_itr)->normal)) < 0)){
// 		  //if((*p_itr)->mark == 1){
// 		  //printf("Point1 : %f %f %f \t Point2: %f %f %f\n",(*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*point_itr)->coords[0], (*point_itr)->coords[1],(*point_itr)->coords[2]);
// 		  //printf("Point1-N : %f %f %f \t Point2-N : %f %f %f\n",(*p_itr)->normal[0],(*p_itr)->normal[1],(*p_itr)->normal[2],(*point_itr)->normal[0], (*point_itr)->normal[1],(*point_itr)->normal[2]);
// 		  //}
// 		  //printf("Co-ordinates are : %f %f %f \t  %f %f %f\n",(*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*point_itr)->coords[0],(*point_itr)->coords[1],(*point_itr)->coords[2] );
// 		  //continue;
// 		  //}
// 		  point_point_visibility(*p_itr,*point_itr,delta);
// 		}
// 		  /*        else   /// start of else loop which considers 2 pts. if visible
// 			    {
// 			    vector<PatchPoint*> Points;
// 			    root->check_intersect(c1,c2, Points);  // the intersection function
// 			    //		printf("Points array size() is: %d\n",Points.size());
// 			    vector<float> dist_array; 
// 			    calc_dist(Points, (*p_itr)->coords, (*point_itr)->coords, delta, dist_array, c1, c2);
// 			    //		calc_min(dist_array, mini, delta)
// 			    for(int in=0;in<CLOSEST;in++)
// 			    {    
// 			    long int arrpos=0;
// 			    for(int itr=0;itr<dist_array.size();itr++)
// 			    {
// 			    if(dist_array[itr] < mini)
// 			    {
// 			    mini = dist_array[itr];
// 			    arrpos = itr;
// 			    } 
// 			    }
// 			    if(mini >= delta)
// 			    break;
// 			    if((mini-0.0) <= 1e-6)
// 			    break;
// 			    if(dist_array.size() > 0){
// 			    dist_array.erase(dist_array.begin()+arrpos);
// 			    }
// 			    double two_pie = 2*22/7;
// 			    double square_root = sqrt(two_pie);
// 			    temp_dist[in] = exp(-(mini*mini)/2)/square_root;
// 			    sum = sum + temp_dist[in];
// 			    }
                                  
// 			    }                                          /// end of else loop which considers 2 pts. if visible
// 			    float total_visible=0;
// 			    if((mini-0.0) <= 1e-6){total_visible = 1; }
// 			    else if(sum != 0){
// 			    for(int in=0;in<CLOSEST;in++)
// 			    total_visible += (temp_dist[in]/sum)*temp_dist[in]; 
// 			    }
// 		  */
// 	      else{ 
// 		if((*p_itr)->mark == 1){
// 		      for(int k=0;k<3;k++)
// 			cell_centers.push_back((*p_itr)->coords[k]);			
// 		      //				printf("Point1 : %f %f %f \t Point2: %f %f %f\n",(*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*point_itr)->coords[0], (*point_itr)->coords[1],(*point_itr)->coords[2]);
// 		      for(int k=0;k<3;k++)
// 			cell_centers.push_back((*point_itr)->coords[k]);
// 		      //				printf("Point1-N : %f %f %f \t Point2-N : %f %f %f\n",(*p_itr)->normal[0],(*p_itr)->normal[1],(*p_itr)->normal[2],(*point_itr)->normal[0], (*point_itr)->normal[1],(*point_itr)->normal[2]);
// 		    }
// 		    total_visible = 0; 
// 		  }
// 		  if((1-total_visible) >= 0.5)
// 		    v_itr=v_itr + 1;
// 		  if(v_itr >= (0.3*(float)(leafcell[j]->patchPoints.size()))){
// 		    visible_list.push_back('1');
// 		    break;
// 		  }
// 		}     /// end of for loop which considers all pts. in a leaf node
// 	      if(visible_list.size() >= ((30.0/100.0)*(float)(leafcell[i]->patchPoints.size()))){
// 		total++;
// 		break;
// 	      }
// 	    }                       /// end of loop which considers all pts. in a leaf node (p_itr)
// 	}
//       // printf("iterator1: %d\n",iterator1);
//     }
//   // printf("total value is : %d \n",total);
//   if(total==(e1-s1+1)*(e2-s2+1))return(1);
//   if(total==0)return(-1);
//   else return(0);


// }                               /// end of function

/*****************************************************************************/
/*****************************************************************************/

bool Cell::operator==(Cell* c2)
{
  if((center[0]-c2->center[0])<= 1e-6){
    if((center[1]-c2->center[1])<= 1e-6){
      if((center[2]-c2->center[2])<= 1e-6){
	return(1);
      }
    }
  }
  else
    return(0);
}

/*****************************************************************************/
/*****************************************************************************/

void Cell::modify()
{
  if(!isLeaf()) {
    for(int j=0;j<init_farCells.size();j++)
      {
	int v;
	v=visi_leaf(start,end,init_farCells[j]->start,init_farCells[j]->end);
	if(v==0) {
	  for(int i=0;i<children.size();i++) {
	    for(int k=0;k<init_farCells[j]->children.size();k++) {
	      children[i]->init_farCells.push_back(init_farCells[j]->children[k]);
	    }}}
	else if(v==1) {
	  farCells.push_back(init_farCells[j]);
	  for(int jj=0;jj<init_farCells[j]->init_farCells.size();jj++){
	    if(init_farCells[j]->init_farCells[jj] == this){
	      init_farCells[j]->farCells.push_back(this);
	      init_farCells[j]->init_farCells.erase(init_farCells[j]->init_farCells.begin()+jj);
	      break;
	    }
	  }
	}
      }
    init_farCells.clear();
    
    for(int j=0;j<children.size();j++) {
      children[j]->modify();}
  }
  else {
    for(int j=0;j<init_farCells.size();j++) {   
      int v=visi_leaf(start,start,init_farCells[j]->start,init_farCells[j]->start);
      if(v==1) {
	farCells.push_back(init_farCells[j]);
	for(int jj=0;jj<init_farCells[j]->init_farCells.size();jj++){
	  if(init_farCells[j]->init_farCells[jj] == this){
	    init_farCells[j]->farCells.push_back(this);
	    init_farCells[j]->init_farCells.erase(init_farCells[j]->init_farCells.begin()+jj);
	    break;
	  }
	}
      }
    }
    init_farCells.clear();

    for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
        if((*itr)->mark == 1){
                cout<<"far cell size is :: " <<farCells.size()<<endl;
        }
    }

    
    for(int j=0;j<init_nearCells.size();j++) {
      int v=visi_leaf(start,start,init_nearCells[j]->start,init_nearCells[j]->start);
      if(v==1) {
	nearCells.push_back(init_nearCells[j]);
	for(int jj=0;jj<init_nearCells[j]->init_nearCells.size();jj++){
	  if(init_nearCells[j]->init_nearCells[jj] == this){
	    init_nearCells[j]->nearCells.push_back(this);
	    init_nearCells[j]->init_nearCells.erase(init_nearCells[j]->init_nearCells.begin()+jj);
	    break;
	  }
	}
      }
    }
    init_nearCells.clear();
  }
}


/*****************************************************************************/
/*****************************************************************************/


void Cell::tofile(FILE* fp_tofile)
{
  if(!isLeaf()) 
    {
      for(int j=0;j<children.size();j++)
	children[j]->tofile(fp_tofile);
    }
  if(isLeaf())
    {
      for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++)
	fprintf(fp_tofile," %f %f %f %f %f %f %f %f %f %f %f %f \n",(*itr)->coords[0], (*itr)->coords[1],(*itr)->coords[2], (*itr)->normal[0],(*itr)->normal[1],(*itr)->normal[2],(*itr)->rhoVal[0], (*itr)->rhoVal[1], (*itr)->rhoVal[2], (*itr)->radVal[0], (*itr)->radVal[1],(*itr)->radVal[2]);
    }
  
}

/*****************************************************************************/
/*****************************************************************************/


//void Cell::get_info(vector<float>& cell_centers)
void Cell::get_info()
{
  if(!isLeaf())
  {
    for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
        if((*itr)->mark == 1){
                cout<<"FARCELL SIZE IS :: "<<farCells.size()<<endl;
                Vec3f v;
                for(int i=0;i<farCells.size();i++){
		    for(int k=farCells[i]->start;k<farCells[i]->end;k++){
                        for(int ii=0;ii<8;ii++){
                                v[0] = leafcell[k]->center[0] + (axesDisplace[ii][0]*leafcell[k]->side);
                                v[1] = leafcell[k]->center[1] + (axesDisplace[ii][1]*leafcell[k]->side);
                                v[2] = leafcell[k]->center[2] + (axesDisplace[ii][2]*leafcell[k]->side);

                                cell_centers.push_back(v);
                        }
		    }
                }
        }
     }  
      for(int j=0;j<children.size();j++)
	children[j]->get_info();
  }
  else
    {
      for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
	if((*itr)->mark == 1){
		cout<<"FARCELL SIZE IS :: "<<farCells.size()<<endl;
		Vec3f v;
		for(int i=0;i<farCells.size();i++){
			for(int ii=0;ii<8;ii++){
                		v[0] = farCells[i]->center[0] + (axesDisplace[ii][0]*farCells[i]->side);
	                	v[1] = farCells[i]->center[1] + (axesDisplace[ii][1]*farCells[i]->side);
	        	        v[2] = farCells[i]->center[2] + (axesDisplace[ii][2]*farCells[i]->side);
		
        	        	cell_centers.push_back(v);
			}
	        }	
	}
     }
    }

}


/*************************************************************************/
/*************************************************************************/

bool Cell::contains(Vec3f pt){   //whether a point is inside cell
  Vec3f diff=pt-center;   // THIS IS NEW UPDATE 6-1-06..ADD ROOT CENTER TO MAKE COORDINATES ABSOLUTE !!
  float xProject=dot(diff,Vec3f(1,0,0));
  if ((fabs(xProject))>(side/2)) return false;
  float yProject=dot(diff,Vec3f(0,1,0));
  if ((fabs(yProject))>(side/2)) return false;
  float zProject=dot(diff,Vec3f(0,0,1));
  if ((fabs(zProject))>(side/2)) return false;
  return true;
}



/*************************************************************************/
/*************************************************************************/

bool Cell::isNearNeighbor(Cell* other){   //check for alternate
  for(int i=0;i<8;i++){
    for(int j=0;j<8;j++){
      if (len((center+axesDisplace[i]*side)-(other->center+axesDisplace[j]*other->side))<=1e-6){
	return true;
      }
    }
  }
  return false;
}




/************************************************************************/

/***********************************************************************************/

void Cell::createlist(){
  if(!parent){
    init_nearCells.push_back(this);
  }
  else {
    vector<Cell*> parentsNeighbors=parent->init_nearCells;
    for(int i=0;i<parentsNeighbors.size();i++){
      for(int j=0;j<parentsNeighbors[i]->children.size();j++){
	if (isNearNeighbor(parentsNeighbors[i]->children[j]))
	  {
	    init_nearCells.push_back(parentsNeighbors[i]->children[j]);
	  }
	else
          {
	    init_farCells.push_back(parentsNeighbors[i]->children[j]);
	  }
      }
    }
  }
  for(int i=0;i<children.size();i++)
    children[i]->createlist();
}

/**************************************************************************/
/**************************************************************************/

int readDepth() {return maxDepth;}

/**************************************************************************/
/**************************************************************************/

void Cell::assgn_weight()
{
  int maxL = readDepth();
  if(!isLeaf()){
    for(int i=0;i<children.size();i++)
      children[i]->assgn_weight();
  }
  else{
    float weight;
    for(PPListItr p_itr1 = patchPoints.begin();p_itr1 != patchPoints.end(); p_itr1++)
      {
	if((*p_itr1)->assgn == -1)
	  {
	    float distance;
	    float close_pts=0.0;
	    Vec3f n1 = (*p_itr1)->normal;
	    //	  for(int j=0;j<nearCells.size();j++){
	    for(PPListItr p_itr2 = patchPoints.begin();p_itr2 != patchPoints.end(); p_itr2++)
	      {
		if((*p_itr2)->assgn == -1)
		  {
		    distance = sqrt( (((*p_itr1)->coords[0])-((*p_itr2)->coords[0]))*(((*p_itr1)->coords[0])-((*p_itr2)->coords[0])) + (((*p_itr1)->coords[1])-((*p_itr2)->coords[1]))*(((*p_itr1)->coords[1])-((*p_itr2)->coords[1])) + (((*p_itr1)->coords[2])-((*p_itr2)->coords[2]))*(((*p_itr1)->coords[2])-((*p_itr2)->coords[2])) );

		    //		    if(distance < (lvl/maxL)*(lvl/maxL)*side/2)
		    if(distance < (depth/maxL)*side/2)
		      {
			Vec3f n2 = (*p_itr2)->normal;
			float theta;
			n1 = norm(n1);
			n2 = norm(n2);
			/*			for(int j=0;j<3;j++)
			  {
			  n1[j] = n1[j]/(sqrt((n1[0]*n1[0]) + (n1[1]*n1[1]) + (n1[2]*n1[2])));
			  n2[j] = n2[j]/(sqrt((n2[0]*n2[0]) + (n2[1]*n2[1]) + (n2[2]*n2[2])));
			  }
			*/
			theta = acosf(dot(n1,n2)) * 180/3.142;
			if(theta < DIFF_ANGLE)
			  {
			    (*p_itr2)->assgn=0;
			    close_pts = close_pts + 1;
			  }
		      }     //end of distance's if loop
		    
		  }       //end of inner assgn's if loop
		
	      }        // end of inner for loop
	    
	    weight = 1/close_pts;
	    for(PPListItr p_itr2 = patchPoints.begin();p_itr2 != patchPoints.end(); p_itr2++)
	      {
		if((*p_itr2)->assgn == 0)
		  {
		    (*p_itr2)->assgn = 1;
		    (*p_itr2)->weight = weight;
		  }
	      }
	    //	}       //end of near Cell loop
	  }	  //end of if assn statement
      }	  // end of outer for loop
  }		  //end of else loop
		   
}

/**************************************************************************************/

/**************************************************************************/

void Cell::createtree(PPList& curPatchPoints,int minpts) { 
  if(depth==0) {patchPoints = curPatchPoints;parent=NULL;nm=0; maxDepth++;}
  start=nm;
  PPListItr itr;
  for(int i=0;i<8;i++){
    Cell* newCell=new Cell(this,center+(axesDisplace[i]*side)/2.0,side/2.0);
    PPList newCellPoints;
    itr=curPatchPoints.begin();
    while(itr!=curPatchPoints.end()){
      if (newCell->contains((*itr)->coords)){
	newCellPoints.push_back((*itr));
	itr=curPatchPoints.erase(itr);
      }
      else{itr++;}
    } 
    if (newCellPoints.size()==0){
      delete newCell;
    }
    else
      {
	if(newCell->depth > maxDepth) {
	  maxDepth = newCell->depth;
	}
	
	newCell->patchPoints = newCellPoints;
	//if(newCell->patchPoints.size() > minpts && side > 0.01){
	if(newCell->side >= CELL_LENGTH){             
	  newCell->createtree(newCellPoints,minpts);}
	else {
	  nm++; newCell->start=nm-1;
	  leafcell.push_back(newCell);
	}
	children.push_back(newCell);
      }
  }
  end=nm-1;
}

long int count_leaf(){return(leafcell.size());}

void Cell::pull(int channel)
{
 
  mecs->reset();

  /*Calculates Multipole Coefficients for every Cell by :
    1)Adding multipole coefficents calculated from patchpoints (for leaf)
    2)By Adding the Multipole - Multipole translated coefficents (for parents) */ 

  if(isLeaf())
    {
      for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
	mecs->addPatchPointMultipole((*itr),channel);
      }
    }
  for(int i=0;i<children.size();i++){
    children[i]->pull(channel);
    mecs->addTranslatedMultipole(children[i]->mecs);
  }
}


/////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef FILE_RENDER
extern ofstream fout;
#endif

/////////////////////////////////////////////////////////////////////////////////////////////////

void Cell::push(int channel, int no_itr)
{
  lecs->reset();
  /*Add Parent's Translated Local Coefficients
    Local - Local Translation */
  if(parent){
    lecs->addTranslatedLocal(parent->lecs);
  }
  /*Add Interaction List Multipole Coefficients
    Multipole - Local Translation */
  for(int i=0;i<farCells.size();i++){
    lecs->addTranslatedMultipole(farCells[i]->mecs);
  }
  /*If leaf then calculate the Radiance Values due to :
    1) Interaction List elements + Parents Local Coefficients
    2) Near Neigbours direct evaluation */
  if(isLeaf()){
    Vec3f newRad;
    for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
      float lecs_value = lecs->evaluateAtPoint((*itr)->coords,(*itr)->normal,(*itr)->rhoVal[channel]);
      if(lecs_value < 0){lecs_value *= -1;}
      newRad[channel]=(*itr)->emmVal[channel]+lecs_value;
      for(int j=0;j<nearCells.size();j++){
	for(PPListItr itr3=nearCells[j]->patchPoints.begin();itr3!=nearCells[j]->patchPoints.end();itr3++){
	  if((*itr)!=(*itr3)) 
	    {
	      float val = (*itr)->evaluateDirect((*itr3),channel);
	      newRad[channel]+=val;
	    }
	}
      }
      (*itr)->newRadVal[channel] += (newRad[channel] - (*itr)->radVal[channel])/(1.3*no_itr); //Convergence step
	
#ifdef FILE_RENDER
      if(channel == 2) {
	fout << (*itr)->coords << (*itr)->normal << (*itr)->newRadVal << RADIUS << endl; 
      }
#endif
    }
  }
     
  else{
    for(int i=0;i<children.size();i++){
      children[i]->push(channel, no_itr);
    }
  }
 
}


void Cell::copy_rad()
{
  if(!isLeaf())
    {
      for(int j=0;j<children.size();j++)
	children[j]->copy_rad();
    }
  if(isLeaf())
    {
      for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++) {
	for(int channel=0;channel<3;channel++) {
	  (*itr)->radVal[channel] = (*itr)->newRadVal[channel];
	}
      }
    }
}
