#include "Radical.hpp"
#include "Cells.hpp"
#include "PointData.hpp"
#include "PatchPoint.hpp"
#include "Mesh.hpp"
//#include "ReadMesh.hpp"
//#include "ReadMaterial.hpp"
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include<assert.h>
#include <signal.h>
#include <fstream>
#include <sys/resource.h>
#include <vector>
#include <list>

/*OGL_RENDER FILE_RENDER DEBUG macros are defined to 
 do conditional compilation of the code and generating
 varying output.

 OGL_RENDER : renders the scene using opengl
 FILE_RENDER : renders the scene into output file "SurfelData". Co-ordinates,
 Normal, Color and Splat radius are output in a file.
 DEBUG : prints all sort of garbage to help in debugging the code.
*/

#ifdef OGL_RENDER
#define RENDER
#elif defined SURFEL_RENDER
#define RENDER
#endif

#ifdef RENDER 
#include <GL/glut.h>
#include <GL/gl.h>

#ifdef OGL_RENDER
//#include "vl/VLgl.h"
#elif defined SURFEL_RENDER
#endif

#endif

#ifdef SURFEL_RENDER
void surfel_render_main();
#endif

#ifdef OGL_RENDER
#define NEAR_CLIP_PLANE 1.0
#define FAR_CLIP_PLANE 100000.0

#define BASE_Z 28
#define BASE_ANG_X 0.0
#define BASE_ANG_Y 0.0

GLfloat eye_z = BASE_Z;
GLfloat rot_x = BASE_ANG_X;
GLfloat rot_y = BASE_ANG_Y;
GLfloat rot_z = 0.0;
#endif

#ifdef FILE_RENDER
ofstream fout("SurfelData");
#endif
int minpts = 10; /*Minimum Points a leaf cell should have. Keep on dividing
		  if the number of points in a cell > minpts*/
Vec3f pointCoord(0,0,0);
Vec3f RootCenter;
vector<Vec3f> cell_centers;
int no_of_itr = 0;
float delta;
PDList surfelList;

#ifdef OGL_RENDER
int screenWidth = 800;     //Height and width of the screen window
int screenHeight = 600;
float windowLeft, windowRight, windowNear, windowFar, windowBottom, windowTop;
float windowWidth = windowRight-windowLeft;
float windowHeight = windowTop-windowBottom;
const float center_wW = windowWidth / 2.0;   //Center of world window
const float center_wH = windowHeight / 2.0;
GLdouble AR = windowWidth / windowHeight;    //Aspect ratio of world window

void normalize(vector<PointData*>&, long int);
void keyboard(unsigned char key,int x,int y);
void display();
void reshape(int w,int h);
void idle();
#endif

vector<PatchPoint*> Qplist;      // list of quadraturepoints
long int numVertices, NoOfLeaf;  // no. of vertices
Point_Mesh *m;                 // pointer to Point_Mesh
//vector<Mesh> pMesh;
//vector<Material> pMtl;
Cell *root;                    // pointer to root cell
int Cell::nm=0;

/*Manage time taken by different modules.*/
struct rusage rstart,rend;
struct timeval start;
struct timeval now;
double second[9];

#ifdef OGL_RENDER
GLfloat eyez=0;
GLfloat eyex=0;
GLfloat eyey=0;
GLfloat rangle=0;
#endif

int readDepth();
long int count_leaf();

/*function for time difference*/
double rdiff(struct rusage *rstart,struct rusage *rend) {
  return rend->ru_utime.tv_sec-rstart->ru_utime.tv_sec+(rend->ru_utime.tv_usec-rstart->ru_utime.tv_usec)*1e-6;
}

/******************************************************/

#ifdef OGL_RENDER

void setWindow(float left, float right, float bottom, float top, float near, float far)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45,1.0,0.0,100.0);
}

#endif

/********* Function for setting the viewport*******************/

#ifdef OGL_RENDER

void setViewport(int left, int right, int bottom, int top)
{
  glViewport(left, bottom, right - left, top - bottom);
}

#endif

/***************** Initialization of the environment*******************/

#ifdef OGL_RENDER

void myInit(void)
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glPointSize(5.0);
  setWindow(windowLeft, windowRight, windowBottom, windowTop,windowNear, windowNear-windowFar);
  setViewport(0, screenWidth, 0, screenHeight);
}

#endif

#ifdef OGL_RENDER

void myKeyboard(unsigned char key, GLint x, GLint y) {
  GLfloat rval = 5.0;
  GLfloat val = .1;
  switch(key) {
  case 's':
    eye_z += val;
    break;
  case 'S':
    eye_z -= val;
    break;
  case 'x':
    rot_x += rval;
    break;
  case 'y':
    rot_y += rval;
    break;
  case 'z':
    rot_z += rval;
     break;
  case 'X':
    rot_x -= rval;
    break;
  case 'Y':
    rot_y -= rval;
    break;
  case 'Z':
    rot_z -= rval;
    break;
  case 'q':
    rot_x = BASE_ANG_X;
    rot_y = BASE_ANG_Y;
    rot_z = 0;
    eye_z = BASE_Z;
    break;
  case 'i':/*******************FMM stuff*********************/
	no_of_itr++;
	surfelList.clear();
        for(int i=0;i<3;i++){
	  cout <<"STARTING THE ITERATION" << endl;
                getrusage(RUSAGE_SELF,&rstart);
                root->pull(i);
		root->push(i,no_of_itr);
		getrusage(RUSAGE_SELF,&rend);
                second[i+6]=rdiff(&rstart,&rend);
		glutPostRedisplay();
        }
	root->copy_pts(surfelList);
	normalize(surfelList, numVertices);
	//for(int channel=0;channel<3;channel++) {
	  root->copy_rad();
	  //}
	break;
  default:
        break;
  }
  glutPostRedisplay();
}

#endif

void Cell::copy_pts(vector<PointData*>& pts)
{
  if(!isLeaf())
    {
      for(int j=0;j<children.size();j++)
	children[j]->copy_pts(pts);
    }
  if(isLeaf())
    {
      Vec3f coords, radVal, normal;
      for(PPListItr itr=patchPoints.begin();itr!=patchPoints.end();itr++){
	coords = (*itr)->coords; radVal = (*itr)->rhoVal; normal = (*itr)->normal;
	PointData* p=new PointData(coords,radVal,normal);
	pts.push_back(p);
      }
    }
}

#ifdef OGL_RENDER

void idle()
{/*glutPostRedisplay();*/
  //Do not do anything if Screen is idle.
}

#endif

#ifdef OGL_RENDER

void drawDot(GLfloat x, GLfloat y, GLfloat z, GLfloat r, GLfloat g, GLfloat b, GLfloat n1, GLfloat n2, GLfloat n3)
{
  glColor3f(r,g,b);
  glBegin(GL_POINTS);
  glVertex3f(x,y,z);
  glEnd();
  
  glBegin(GL_LINES);
  glVertex3f(x,y,z);
  glVertex3f(x+n1/2,y+n2/2,z+n3/2);
  glEnd();
}

#endif

#ifdef RENDER

void myDisplay(){

#ifdef OGL_RENDER
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glClearDepth(0.0);
  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  gluLookAt(0.0,0.0,eye_z+5,0.0,0.0,0.0,0.0,1.0,0.0);
  glRotatef(rot_x,1,0,0);
  glRotatef(rot_y,0,1,0);
  glRotatef(rot_z,0,0,1);
  glViewport(0,0,800,600);
  for(PDListItr p_itr = surfelList.begin();p_itr != surfelList.end();p_itr++)
    {
      drawDot((*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*p_itr)->radVal[0],(*p_itr)->radVal[1],(*p_itr)->radVal[2],(*p_itr)->normal[0],(*p_itr)->normal[1],(*p_itr)->normal[2]);
    }
  glPointSize(1.0);
  glColor3f(1.0f,1.0f,0.0f);
/*  for(int i=0;i<cell_centers.size()-1;i+=2){     //drawing Yellow lines
     glBegin(GL_LINES);
     glVertex3f(cell_centers[i][0],cell_centers[i][1],cell_centers[i][2]);
     glVertex3f(cell_centers[i+1][0],cell_centers[i+1][1],cell_centers[i+1][2]);
     glEnd();
   }
*/
/************ for visible cells *************/
   
   cout<< "CELL CENTER SIZE() IS :: " <<cell_centers.size()<<endl; 
   for(int i=0;i<=cell_centers.size()-8;i+=8){
                glBegin(GL_LINE_LOOP);
                        glVertex3f(cell_centers[i][0],cell_centers[i][1],cell_centers[i][2]);
                        glVertex3f(cell_centers[i+1][0],cell_centers[i+1][1],cell_centers[i+1][2]);
                        glVertex3f(cell_centers[i+2][0],cell_centers[i+2][1],cell_centers[i+2][2]);
                        glVertex3f(cell_centers[i+3][0],cell_centers[i+3][1],cell_centers[i+3][2]);
			
                glEnd();
                glBegin(GL_LINE_LOOP);
			glVertex3f(cell_centers[i+2][0],cell_centers[i+2][1],cell_centers[i+2][2]);
                        glVertex3f(cell_centers[i+3][0],cell_centers[i+3][1],cell_centers[i+3][2]);
			glVertex3f(cell_centers[i+4][0],cell_centers[i+4][1],cell_centers[i+4][2]);
                        glVertex3f(cell_centers[i+5][0],cell_centers[i+5][1],cell_centers[i+5][2]);
                glEnd();
                glBegin(GL_LINE_LOOP);
			glVertex3f(cell_centers[i+4][0],cell_centers[i+4][1],cell_centers[i+4][2]);
                        glVertex3f(cell_centers[i+5][0],cell_centers[i+5][1],cell_centers[i+5][2]);
			glVertex3f(cell_centers[i+6][0],cell_centers[i+6][1],cell_centers[i+6][2]);
                        glVertex3f(cell_centers[i+7][0],cell_centers[i+7][1],cell_centers[i+7][2]);
                glEnd();
                glBegin(GL_LINE_LOOP);
			glVertex3f(cell_centers[i+6][0],cell_centers[i+6][1],cell_centers[i+6][2]);
                        glVertex3f(cell_centers[i+7][0],cell_centers[i+7][1],cell_centers[i+7][2]);
			glVertex3f(cell_centers[i][0],cell_centers[i][1],cell_centers[i][2]);
                        glVertex3f(cell_centers[i+1][0],cell_centers[i+1][1],cell_centers[i+1][2]);			
                glEnd();
        }	
   glPointSize(10.0);
   glColor3f(0.0f,1.0f,1.0f);
   cout<<"Point is :: "<<pointCoord<<endl;
   glBegin(GL_POINTS);
   glVertex3f(pointCoord[0],pointCoord[1],pointCoord[2]);
   glEnd();
   glPointSize(5.0);
   glPopMatrix();
   glutSwapBuffers();
   glFlush();

#elif defined SURFEL_RENDER
   /*This is the section where point primitives are rendered using surfel splatting*/
   //surfel_render_main();
#endif
}

#endif

/************************ END OF mydisplay() ****************************************/

#ifdef OGL_RENDER

void normalize(vector<PointData*>& pts,long int Vertices)
{
  Vec3f* colors=new Vec3f[Vertices];
  Vec3f* points=new Vec3f[Vertices];
  Vec3f pMax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL), pMin(HUGE_VAL,HUGE_VAL,HUGE_VAL);
  Vec3f cMax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
  unsigned long int i=0;
  for(PDListItr p_itr = pts.begin();p_itr!=pts.end();p_itr++){
    for(int j=0;j<3;j++){
      colors[i][j] = (*p_itr)->radVal[j];
      cMax[j]=max(colors[i][j],cMax[j]);
      points[i][j] = (*p_itr)->coords[j];
      pMax[j]=max(points[i][j],pMax[j]);
      pMin[j]=min(points[i][j],pMin[j]);
    }
    i++;
  }
  windowLeft  = pMin[0];
  windowRight = pMax[0];
  windowBottom = pMin[1];
  windowTop = pMax[1];
  windowNear = pMax[2];
  windowFar = pMin[2];

}

#endif

#ifdef OGL_RENDER

void myReshape(GLsizei W, GLsizei H)
{
  int WiS, WiE, HeS, HeE;

  if (AR > W / H)
    setViewport(0, W, 0, int(W/AR));
  else
    setViewport(0, int(H*AR), 0, H);
  glutPostRedisplay();
}

#endif

/***********************************************************************************************************/

/***********************************************************************************************************/

/* main starts*/
int main(int argc,char **argv)
{
  minpts = atoi(argv[3]);
  initMath(5);                      //initialise math
  delta = atof(argv[4]);
  delta = sqrt(delta)*(atof(argv[5]));
  
  int flag = atoi(argv[1]);
  if(flag){
    //ReadMesh(pMesh, pMtl, argv[2]);
    //Subdivide(pMesh, pMtl);
    //numVertices = populatePatchPoints(Qplist, pMesh);
  }
  else {
    getrusage(RUSAGE_SELF,&rstart);
    numVertices = createPatchPoints(Qplist, argv[2]); 
    getrusage(RUSAGE_SELF,&rend);
    second[0] =  rdiff(&rstart,&rend);
  }
  

  /****************************************************************************/ 

  getrusage(RUSAGE_SELF,&rstart);
  m = new Point_Mesh(Qplist, numVertices);    //This changes the co-ordinates of the points too !!
#ifdef DEBUG
  FILE *fp = fopen("point_rendering","w");
  for(PPListItr p_itr = Qplist.begin();p_itr!=Qplist.end();p_itr++){
	fprintf(fp,"%f %f %f %f %f %f %f %f %f %f \n",(*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*p_itr)->normal[0],(*p_itr)->normal[1],(*p_itr)->normal[2],(*p_itr)->diffVal[0],(*p_itr)->diffVal[1],(*p_itr)->diffVal[2],0.5);
  }
  fclose(fp);
#endif
  getrusage(RUSAGE_SELF,&rend);
  second[1] =  rdiff(&rstart,&rend);
  RootCenter = m->center; //THIS IS THE NEW UPDATE !!! 
  float Root_Side = m->side;

  /********create root node**************/

  root=new Cell(NULL,m->center,m->side);
  getrusage(RUSAGE_SELF,&rstart);
  /********create octree structure*******/
  root->createtree(Qplist,minpts);
  getrusage(RUSAGE_SELF,&rend);
  second[2] =  rdiff(&rstart,&rend);
  
  NoOfLeaf = count_leaf();
  cout << "Number of leaf Nodes :: " << NoOfLeaf << endl;
  cout << "Maximum Depth :: " << readDepth() << endl;
  getrusage(RUSAGE_SELF,&rstart);
  root->assgn_weight();
  getrusage(RUSAGE_SELF,&rend);
  second[3] = rdiff(&rstart,&rend);
  
  /********create interaction lists*******/
  getrusage(RUSAGE_SELF,&rstart);
  root->createlist();
  getrusage(RUSAGE_SELF,&rend);
  second[4]=rdiff(&rstart,&rend);
  
  getrusage(RUSAGE_SELF,&rstart);
  /********modify list -- check for visibility***********/
  //root->modify();        //pass the sorted array of X corrd as argument
  getrusage(RUSAGE_SELF,&rend);
  second[5]=rdiff(&rstart,&rend);

  root->get_info();
  cout<<"Cell_center.size() :: "<<cell_centers.size()<<endl;
  no_of_itr++;
  /*for(int i=0;i<3;i++){
    cout <<"STARTING THE ITERATION" << endl;
    getrusage(RUSAGE_SELF,&rstart);
    root->pull(i);
    root->push(i, no_of_itr);
    getrusage(RUSAGE_SELF,&rend);
    second[i+6]=rdiff(&rstart,&rend);
    }*/
  
  root->copy_pts(surfelList);
  
#ifdef OGL_RENDER
  normalize(surfelList, numVertices);
#endif
  //for(int channel=0;channel<3;channel++)
  root->copy_rad();
//    root->get_info();

  cout << "Create Patch Point Time :: " << second[0] << endl;
  cout << "Point Mesh Creation Time :: " << second[1] << endl;
  cout << "Tree Creation Time :: " << second[2] << endl;
  cout << "Assign Weight Time :: " << second[3] << endl;
  cout << "Create List Time :: " << second[4] << endl;
  cout << "Modification Time :: " << second[5] << endl;
  cout << "FMM Push Pull Time :: "<<(second[6]+second[7]+second[8]) << endl;
  
  /********OpenGL initalisation****************/
#ifdef OGL_RENDER
  glutInit(&argc,argv);
  glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE| GLUT_RGB);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH);
  glEnable(GL_DEPTH_TEST);
  glutInitWindowSize(screenWidth,screenHeight);
  glutCreateWindow("FMM PBR - ITERATIONS");
  glutReshapeFunc(myReshape);
  glutKeyboardFunc(myKeyboard);
  glutDisplayFunc(myDisplay);
  myInit();
  glutMainLoop();
#endif

#ifdef SURFEL_RENDER
  cout << "Calling Surfel Render Main" << endl;
  surfel_render_main();
  cout << "Done Surfel Render Main" << endl;
#endif

  /*******************************************/
}
