#include "ssVector.hpp"
#include "ssMatrix.hpp"
#include "ssAdditional.hpp"
#include "ABuffer.h"
#include "struct.h"
#include "def.h"
#include <fstream.h>
#include <iostream>
#include <GL/gl.h>
#include <GL/glut.h>
#include <cstdlib>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <time.h>
#include <sys/resource.h>

#include "Cells.hpp"     
#include "PointData.hpp"
#include "PatchPoint.hpp"

/*keys:

r -- reset camera position

q,w,a,s,z,x = rotate camera
arrows = move camera sideways
1,2 = move camera backwards/forwards
3,4 = rotate object
5,6 = subpixels off/on (stochastic pipeline only)
7,8 = lower/higher resolution
9,0 = zoom the screen
c,v = stochastic/analytic pipeline
b,n,m = normal mode/point visualization/reconstruction visualization
h,j = z-threshold/z-range surface reconstruction
k,l = decrease/increase the threshold of the z-threshold algorithm
F5,F9 = save/load camera position
ESC,ALT+F4 = exit
*/

//4.8 2.0 -2.8 0.0 2.0 0.0 0.0 1.0 0.0

using namespace std;

#define MAX_SPLAT_SIZE .3
#define NBD_SIZE 3

#define PROJECTION_PERSPECTIVE


#define OBJECT_MODEL
//#define OBJECT_ONESPLAT
//#define OBJECT_CHESSBOARD

Color				*buffer = NULL;
ABuffer				*abuffer = NULL;

float				width = defaultWidth;
float				height = defaultHeight;
float cam_x = 0.0;
float cam_y = 0.0;
float cam_z = -25.0;

char kbdBuffer[256];

std::vector<Splat>	splats;

double rdiff(struct rusage *rstart,struct rusage *rend);

extern Cell* root;
extern PDList surfelList;
extern int no_of_itr;
extern struct rusage rstart,rend;
extern struct timeval start;
extern struct timeval now;
extern double second[9];

void mainloop();

void poissonDiscDistribution( Vector2 *points, int count ) {
  float threshold = 1.0f;
  
  //repeat until we have all subpixels distributed
  int j=0;
  while(j < count) {
    //try to place a random sample in a unit circle so that it is not closer
    // to earlier samples than the threshold 
    int tries = 100;
    int k;
    Vector2 p;
    do {
      //make a random sample inside unit circle
      p = randInCircle();
      
      //test whether the sample is closer to points already placed than the threshold
      for(k=0;k<j;k++)
	{
	      if( (p - points[k]).length() < threshold )
		{	//too close, break
		  tries--;
		  break;
		}
	    }
	} while(k < j && tries > 0);
      
      if( tries == 0 )
	{
	  //if did not succeed, lower the threshold and redistribute all points
	  threshold *= 0.98f;	//the closer the number is to 1, the slower the convergence
	  j = 0;
	}
      else
	{
	  //if succeeded, add the sample
	  points[j] = p;
	  j++;
	}
    }
}

float cumulativeGaussian[gausssamples];
  
void initCumulativeGaussian()
{
//init cumulative gaussian
float x = 0.0f;
float xstep = prefiltercutoff / (float)gausssamples;
x += xstep/2.0f;	//calculate x at the center of each slice
cumulativeGaussian[0] = exp(-0.5f*x*x)*xstep;
for(int i=1;i<gausssamples;i++)
{
  x += xstep;
  cumulativeGaussian[i] = cumulativeGaussian[i-1] + exp(-0.5f*x*x)*xstep;
}
//normalize
 for(int i=0;i<gausssamples;i++)
   {
     cumulativeGaussian[i] /= cumulativeGaussian[gausssamples-1];
   }
}

float inverseCumulativeGaussian( float r )
{
  int i;
  for(i=0;i<gausssamples-1;i++)
    {
      if( r > cumulativeGaussian[i] && r < cumulativeGaussian[i+1] )
	break;
    }
  return (float)i/(float)gausssamples;
}

void makeJitter(int subpixels)
{
  //SS_ASSERT(subpixels > 0);
  
  if( prefilterJitter ) delete[] prefilterJitter;
  if( lensJitter) delete[] lensJitter;
  
  prefilterJitter = new Vector2[subpixels*jitterPatterns];
  lensJitter = new Vector2[subpixels*jitterPatterns];


  initCumulativeGaussian();
  for(int p=0;p<jitterPatterns;p++)
    {
      poissonDiscDistribution(&prefilterJitter[p*subpixels],subpixels);
      for(int j=0;j<subpixels;j++)
	{
	  Vector2 jit = prefilterJitter[p*subpixels+j];
	  float r = jit.length();
	  if( r > 0.0f ) jit *= 1.0f/r;
	  jit *= inverseCumulativeGaussian(r) * prefiltersize;
	  prefilterJitter[p*subpixels+j] = jit;
	}
      
      poissonDiscDistribution(&lensJitter[p*subpixels],subpixels);
      for(int j=0;j<subpixels;j++)
	{
	  lensJitter[p*subpixels+j] *= lensRadius;
	}
    }
}

void resizeABuffer(int width, int height, int subpixels)
{
	if( width < 1 || width > maxResolution || height < 1 || height > maxResolution || subpixels < 1 || subpixels > 64 ) return;
	if( abuffer )
	{
		if( width != abuffer->getWidth() || height != abuffer->getHeight() || subpixels != abuffer->getSubpixels() )
		{
			delete abuffer;
			delete buffer;
		}
		else return;
	}
	///if( subpixels != abuffer->getSubpixels() ) makeJitter(subpixels);
	abuffer = new ABuffer(width, height, subpixels );
	buffer = new Color[width * height];
}

void kbdfunc(unsigned char key, int x, int y)
{
  kbdBuffer[key] = 1;
  
  if( key == 27 )
    exit(0);	//esc or alt-F4 exits, TODO make exit so that destructors are called
  if( key == '0' ) pixelScale *= 2.0f;
  if( key == '9' ) pixelScale *= 0.5f;
  if( key == '8' && sizex < maxResolution && sizey < maxResolution)
    {
      sizex *= 2;
      sizey *= 2;
      resizeABuffer(sizex,sizey,subpixels);
    }
  if( key == '7' && sizex > 2 && sizey > 2)
    {
      sizex /= 2;
      sizey /= 2;
      resizeABuffer(sizex,sizey,subpixels);
    }
  if( key == '5' && subpixels > 1 )
    {
      subpixels = 1;
      prefiltersize = 0.0f;
      resizeABuffer(sizex,sizey,subpixels);
    }
  if( key == '6' && subpixels == 1 )
    {
      subpixels = maxSubpixels;
      prefiltersize = maxPrefiltersize;
      resizeABuffer(sizex,sizey,subpixels);
    }
  if( key == 'b' )
    if( rasterization ) rasterization = false;
    else rasterization = true;
  if( key == 'n' )
    if( pointvisualization ) pointvisualization = false;
    else pointvisualization = true;
  if( key == 'm' )
    if( reconstructionvisualization ) reconstructionvisualization = false;
    else reconstructionvisualization = true;
  
  if (key == 'i') {
    no_of_itr++;
    surfelList.clear();
    for(int i=0;i<3;i++){
      cout <<"STARTING THE ITERATION" << endl;
      getrusage(RUSAGE_SELF,&rstart);
      root->pull(i);
      root->push(i,no_of_itr);
      getrusage(RUSAGE_SELF,&rend);
      second[i+6]=rdiff(&rstart,&rend);
      glutPostRedisplay();
    }
    root->copy_pts(surfelList);
    int i=0;
    for(PDListItr p_itr = surfelList.begin();p_itr != surfelList.end();p_itr++,i++) {
      Vector3 coord((*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2]);
      Vector3 normals((*p_itr)->normal[0],(*p_itr)->normal[1],(*p_itr)->normal[2]);
      Vector4 radVal((*p_itr)->radVal[0],(*p_itr)->radVal[1],(*p_itr)->radVal[2],1.0);
      (splats[i]).m_position[0] = coord[0];
      (splats[i]).m_normal[0] = normals[0];
      (splats[i]).m_color[0] = radVal[0];
    }
    //for(int channel=0;channel<3;channel++) {
    root->copy_rad();
    //}
  }
  
  glutPostRedisplay();
}

void kbdupfunc(unsigned char key, int x, int y) { kbdBuffer[key] = 0; }

void specialfunc(int key, int x, int y)
{
  if( key == GLUT_KEY_F4 && glutGetModifiers() == GLUT_ACTIVE_ALT )
    exit(0);	//esc or alt-F4 exits, TODO make exit so that destructors are called
  kbdBuffer[key] = 1;
}

void specialupfunc(int key, int x, int y) { kbdBuffer[key] = 0; }
void idlefunc() { //glutPostRedisplay();
}
void reshape( int w, int h )
{
  width = w;
  height = h;
  glViewport( 0, 0, w, h );
  //glutSetWindowTitle(STR("Super 3D Virtual Reality  (%d, %d)",int(width),int(height)));
}

//============================================================================
//============================================================================

class CameraControls
{
public:
  CameraControls( float rotationScale = 1.0f, float translationScale = 1.0f, float inertia = 0.0f, bool framerateDependent = false )
  {
    m_rotationScale = rotationScale;
    m_translationScale = translationScale;
    m_inertia = inertia;
    m_framerateDependent = framerateDependent;
    
    m_angles.set(0,0,0);
    m_angleAdds.set(0,0,0);
    m_position.set(cam_x,cam_y,cam_z);
    m_positionAdd.set(0,0,0);
    m_objectAngles.set(0,0,0);
    m_cameraMatrix.ident();
  }
  ~CameraControls()
  {
  }
  
  const Matrix4x4 &getCameraMatrix( char *kbdBuffer )
  {
    float s,r;
    //m_timer.sample();
    //float dt = m_timer.getDelta();
    
    //camera movement
    if( m_framerateDependent )
      {
	//s = 5.0f * m_translationScale * dt;
	//	r = m_rotationScale * 3.1415f / 4.0f * dt;
      }
    else
      {
	//s = 0.5f * m_translationScale;
	s = 2.0f * m_translationScale;
	r = m_rotationScale * 3.1415f / 64.0f;
      }
    
    if(kbdBuffer['r']) m_position.set(cam_x,cam_y,cam_z);
    if(kbdBuffer['q']) m_angleAdds.x += r;
    if(kbdBuffer['w']) m_angleAdds.x -= r;
    if(kbdBuffer['a']) m_angleAdds.y += r;
    if(kbdBuffer['s']) m_angleAdds.y -= r;
    if(kbdBuffer['z']) m_angleAdds.z += r;
    if(kbdBuffer['x']) m_angleAdds.z -= r;
    
    if(kbdBuffer[GLUT_KEY_LEFT])
      {
	m_positionAdd.x -= m_cameraMatrix[0][0] * s;
	m_positionAdd.y -= m_cameraMatrix[1][0] * s;
	m_positionAdd.z -= m_cameraMatrix[2][0] * s;
      }
    if(kbdBuffer[GLUT_KEY_RIGHT])
      {
	m_positionAdd.x += m_cameraMatrix[0][0] * s;
	m_positionAdd.y += m_cameraMatrix[1][0] * s;
	m_positionAdd.z += m_cameraMatrix[2][0] * s;
      }
    if(kbdBuffer[GLUT_KEY_DOWN])
      {
	m_positionAdd.x -= m_cameraMatrix[0][1] * s;
	m_positionAdd.y -= m_cameraMatrix[1][1] * s;
	m_positionAdd.z -= m_cameraMatrix[2][1] * s;
      }
    if(kbdBuffer[GLUT_KEY_UP])
      {
	m_positionAdd.x += m_cameraMatrix[0][1] * s;
	m_positionAdd.y += m_cameraMatrix[1][1] * s;
	m_positionAdd.z += m_cameraMatrix[2][1] * s;
      }
    if(kbdBuffer['1'])
      {
	m_positionAdd.x -= m_cameraMatrix[0][2] * s;
	m_positionAdd.y -= m_cameraMatrix[1][2] * s;
	m_positionAdd.z -= m_cameraMatrix[2][2] * s;
      }
    if(kbdBuffer['2'])
      {
	m_positionAdd.x += m_cameraMatrix[0][2] * s;
	m_positionAdd.y += m_cameraMatrix[1][2] * s;
	m_positionAdd.z += m_cameraMatrix[2][2] * s;
      }
    m_angles += m_angleAdds;
    m_position += m_positionAdd;
    if( m_framerateDependent )
      {
	///m_angleAdds *= float(pow(m_inertia,dt*30.0f));
	///m_positionAdd *= float(pow(m_inertia,dt*30.0f));
      }
    else
      {
	m_angleAdds *= m_inertia;
	m_positionAdd *= m_inertia;
      }
    
    if(kbdBuffer[GLUT_KEY_F5])
      {
	std::ofstream o;
	o.open("cameramatrix.txt");
	o << m_angles.x << " " << m_angles.y << " " << m_angles.z << " ";
	o << m_position.x << " " << m_position.y << " " << m_position.z << " ";
	o << (int)width << " " << (int)height << " ";
	o.close();
      }
    if(kbdBuffer[GLUT_KEY_F9])
      {
	std::ifstream o;
	o.open("cameramatrix.txt");
	o >> m_angles.x >> m_angles.y >> m_angles.z;
	o >> m_position.x >> m_position.y >> m_position.z;
	o >> width >> height;
	o.close();
	glutReshapeWindow((int)width,(int)height);
      }
    
    m_cameraMatrix.ident();
    m_cameraMatrix *= MatrixFactory::rotateZ( -m_angles.z );
    m_cameraMatrix *= MatrixFactory::rotateX( -m_angles.x );
    m_cameraMatrix *= MatrixFactory::rotateY( -m_angles.y );
    m_cameraMatrix.setColumn(3,Vector4(m_position.x,m_position.y,m_position.z,1.0f));
    return m_cameraMatrix;
  }
private:
  bool			m_framerateDependent;
  float			m_inertia;
  float			m_rotationScale;
  float			m_translationScale;
  Vector3			m_angles;
  Vector3			m_angleAdds;
  Vector3			m_position;
  Vector3			m_positionAdd;
  Matrix4x4		m_cameraMatrix;
  //	Timer			m_timer;

  Vector3		m_objectAngles;
};


void analyticPipeline(const std::vector<Splat> &splats, ABuffer *abuffer,      const Matrix4x4 &objectToCamera,
		      const Matrix4x4 &projection,
		      const Vector4 &viewport,
		      float front,
		      float back ) {
  
  const int exptablesize = 256;
  static float	exptable[exptablesize];
  static bool	exptableinitialized = false;
  const float	exptablescale = float(exptablesize)/F;
  const float	oosqrtF = 1.0f / sqrt(F);

  if( !exptableinitialized )
    {
      for(int i=0;i<exptablesize;++i)
	{
	  exptable[i] = exp(-0.5f * F * float(i)/float(exptablesize));
	}
    }

  int width = abuffer->getWidth();
  int height = abuffer->getHeight();
  int subpixels = abuffer->getSubpixels();

  int scissorxmin = intFloor(viewport.x);
  int scissorymin = intFloor(viewport.y);
  int scissorxmax = intFloor(viewport.z);
  int scissorymax = intFloor(viewport.w);
  if( scissorxmin < 0 ) scissorxmin = 0;
  if( scissorymin < 0 ) scissorymin = 0;
  if( scissorxmax > width ) scissorxmax = width;
  if( scissorymax > height ) scissorymax = height;

  Vector2 vpScale( 0.5f * (viewport.z - viewport.x), 0.5f * (viewport.w - viewport.y) );
  Vector2 vpMid( 0.5f * (viewport.z + viewport.x), 0.5f * (viewport.w + viewport.y) );
  
  Vector2 oovpScale(1.0f/vpScale.x, 1.0f/vpScale.y);
  
  //prefilter
  Matrix2x2 Vp( prefiltersize*oovpScale.x*oovpScale.x, 0,
		0, prefiltersize*oovpScale.y*oovpScale.y );
  
  Matrix4x4 objectToScreen;
  objectToScreen = objectToCamera;
  objectToScreen *= projection;
  
  Matrix4x4 screenToObject;
  screenToObject = objectToScreen;
  bool invertible = screenToObject.invert();
  
  ///SS_ASSERT(invertible);
  /**/
  if(!invertible) {
    exit(-1);
  }
  
  if( screenToObject[2].w != 0.0f ) screenToObject[2] *= 1.0f/screenToObject[2].w;
  else screenToObject[2] = -screenToObject[2];
  Vector4 c = screenToObject[2];
  
  //Vector4 light(10,10,-20,0);
  Matrix4x4 cameraToObject = objectToCamera;
  cameraToObject.invert();
  //light *= cameraToObject;
  //Vector3 lightInObject(light.x, light.y, light.z);
  //lightInObject.normalize();
  
  //clip plane setup
  Vector4 clipOrigin[6],clipNormal[6];
  int clipPlanes = 6;
  {
    float s = (float)sqrt(2.0)/2.0f;
    clipOrigin[0].set(0,0,0,0);
    clipNormal[0].set(0,0,1,0);
    
    clipOrigin[1].set(0,0,back,0);
    clipNormal[1].set(0,0,-1,0);
    
    clipOrigin[2].set(0,0,0,0);
    clipNormal[2].set(-s,0,0,s);
    
    clipOrigin[3].set(0,0,0,0);
    clipNormal[3].set(s,0,0,s);
    
    clipOrigin[4].set(0,0,0,0);
    clipNormal[4].set(0,-s,0,s);
    
    clipOrigin[5].set(0,0,0,0);
    clipNormal[5].set(0,s,0,s);
  }
  
  int subpixelsTested = 0;
  int subpixelsPassed = 0;
  
  for(int i=0;i<splats.size();++i)
    {
      Vector3 p0 = splats[i].m_position;
      Vector3 n = splats[i].m_normal;
      
      //backface cull (in object-space)
      Vector4 pleq(n.x, n.y, n.z, -dot(p0,n));
      int dotSign = extractSign( dot(c, pleq) );
      if( dotSign ) continue;

      //set up the conic matrix
      Matrix2x2 Vr( splats[i].m_size, 0,
					  0,                splats[i].m_size );
      Matrix2x2 VrT = Vr;
      VrT.transpose();
      Matrix2x2 Q = Vr * VrT;
      Q.invert();

      //make the tangent plane (in object-space)
      Vector3 tu = perpendicular( n );
      tu.normalize();
      Vector3 tv = cross(n, tu);

      //transform the tangents and
      // the splat centre to the screen-space
      Vector4 tuh(tu.x, tu.y, tu.z, 0);
      Vector4 tvh(tv.x, tv.y, tv.z, 0);
      Vector4 p0h(p0.x, p0.y, p0.z, 1);
      tuh *= objectToScreen;
      tvh *= objectToScreen;
      p0h *= objectToScreen;

      //viewport culling
      int cp;
      for(cp=0;cp<clipPlanes;cp++) {
	float a = dot(tuh, clipNormal[cp]);
	float b = dot(tvh, clipNormal[cp]);
	float c = dot(p0h, clipNormal[cp]) - dot(clipOrigin[cp], clipNormal[cp]);
	if( a == 0.0f && b == 0.0f ) { //the planes are parallel
	  if( c < 0.0f ) break; //behind the clip plane
	} else {
	  if( F * (Q[0][0]*b*b - 2.0f*Q[0][1]*a*b + Q[1][1]*a*a) - c*c*Q.det() >= 0.0f )
	    { //the clip plane intersects the splat
	      if( cp == 0 ) break; //front plane discards the splat
	    } else { //the clip plane doesn't intersect the splat
	      if( c < 0.0f ) break; //behind the clip plane
	    }
	}
      }
      if( cp < clipPlanes ) continue;	//the splat was behind some of the planes
      
      //mapping from the splat to the screen-space
      Matrix3x3 M(tuh.x, tuh.y, tuh.w,
		  tvh.x, tvh.y, tvh.w,
		  p0h.x, p0h.y, p0h.w);
      
      //mapping from the screen-space to the splat
      if( M.det() == 0.0f ) { cout << "Matrix Determinant is Zero :( " << endl; continue;} //don't draw if the matrix is singular
      Matrix3x3 Mi = M;
      Mi.invert();
      
      //discard the splat if the mapping matrix's condition number is too high
      float conditionNumber = norm_inf(M) * norm_inf(Mi);
      //if( conditionNumber > conditionThreshold ) {cout << "Condition Number is High :( " << endl; continue; }
      
      //affine approximation
      Matrix3x3 Qh( Q[0][0], Q[0][1], 0,
		    Q[1][0], Q[1][1], 0,
		    0,       0,       -F );
      Matrix3x3 MiT = Mi;
      MiT.transpose();
      Matrix3x3 Qhs = Mi * Qh * MiT;
      
      float delta = Qhs[0][0] * Qhs[1][1] - Qhs[0][1] * Qhs[1][0];
      if( delta <= 0.0f ) continue; //draw only ellipses
      float oodelta = 1.0f/delta;
      
      float xt = (Qhs[0][1] * Qhs[1][2] - Qhs[1][1] * Qhs[0][2]) * oodelta;
      float yt = (Qhs[0][1] * Qhs[0][2] - Qhs[0][0] * Qhs[1][2]) * oodelta;
      
      Matrix2x2 Qs( Qhs[0][0], Qhs[0][1],
		    Qhs[1][0], Qhs[1][1] );
      
      float f = -Qhs[2][2] - Qhs[0][2] * xt - Qhs[1][2] * yt;
      if( f <= 0.0f ) continue;
      Qs *= F / f;
      
      //make the resampling filter, the code below is equivalent to
      // Qs = Invert( Invert(Qs) + Vp );
      float detQs = Qs.det();
      if( detQs <= 0.0f ) continue;
      Qs[0][0] = (Qs[0][0] + detQs*Vp[1][1]);
      Qs[0][1] = (Qs[0][1] - detQs*Vp[0][1]);
      Qs[1][0] = (Qs[1][0] - detQs*Vp[1][0]);
      Qs[1][1] = (Qs[1][1] + detQs*Vp[0][0]);
      Qs *= detQs/Qs.det();
      
      //the bounding rectangle
      delta = Qs.det();
      if( delta <= 0.0f ) continue;
      oodelta = 1.0f / delta;
      float bx = sqrt(Qs[1][1] * F * oodelta);
      float by = sqrt(Qs[0][0] * F * oodelta);
      float xmin = xt - bx, xmax = xt + bx;
      float ymin = yt - by, ymax = yt + by;
      
      //determine the integer bounding rectangle and do scissoring
      xmin = xmin * vpScale.x + vpMid.x;
      xmax = xmax * vpScale.x + vpMid.x;
      ymin = ymin * vpScale.y + vpMid.y;
      ymax = ymax * vpScale.y + vpMid.y;
      
      int sx = intFloor( xmin ), ex = intCeil( xmax );
      if( sx >= width || ex <= 0 ) continue;
      sx = max2(sx,scissorxmin);
      ex = min2(ex,scissorxmax);
      if( ex - sx == 0 ) continue;
      
      int sy = intFloor( ymin ), ey = intCeil( ymax );
      if( sy >= height || ey <= 0 ) continue;
      sy = max2(sy,scissorymin);
      ey = min2(ey,scissorymax);
      if( ey - sy == 0 ) continue;
      
      //depth interpolation setup
      float zmin = p0h.z,zmax = p0h.z;
      float zx = 0.0f,zy = 0.0f;
      float zt = p0h.z;
      float l = Q[0][0]*tuh.z*tuh.z + 2.0f*Q[0][1]*tuh.z*tvh.z + Q[1][1]*tvh.z*tvh.z;
      if( l != 0.0f ) {
	//calculate zmin and zmax
	l = sqrt(max2(0.0f,F/l));
	zmin = p0h.z - l * (tuh.z*tuh.z + tvh.z*tvh.z);
	zmax = p0h.z + l * (tuh.z*tuh.z + tvh.z*tvh.z);
	zt = (zmin + zmax)*0.5f;
	
	//calculate the depth gradient, not needed for our surface reconstruction
	zx = -tvh.z * M[0][1] + tuh.z * M[1][1];
	zy =  tvh.z * M[0][0] - tuh.z * M[1][0];
	float dz = F*(zy*zy*Qs[0][0] - 2*zy*zx*Qs[0][1] + zx*zx*Qs[1][1]);
	if( dz != 0.0f ) dz = (zmax-zt) * sqrt(max2(0.0f,delta / dz));
	zx *= dz;
	zy *= dz;
      }
      
      //call the splat shader here
      Vector4 color(0,0,0,1);
      color = splats[i].m_color;
      
      if( rasterization )
	{
	  //rasterization loop
	  ABuffer::Fragment fragment;
	  fragment.zmin = zmin;
	  fragment.zmax = zmax;
	  fragment.dotSign = dotSign;
	  fragment.color = color;
	  for(int y=sy;y<ey;++y)
	    {
	      for(int x=sx;x<ex;++x)
		{
		  Vector2 xp(((float)x - vpMid.x)*oovpScale.x - xt,
			     ((float)y - vpMid.y)*oovpScale.y - yt);
		  subpixelsTested++;
		  float radius = Qs[0][0]*xp.x*xp.x + 2.0f*Qs[0][1]*xp.x*xp.y + Qs[1][1]*xp.y*xp.y;
		  if( radius < F )
		    {
		      subpixelsPassed++;
		      float z = zt + xp.x * zx + xp.y * zy; //needed, if the surface reconstruction uses z
		      fragment.weight = exptable[max2(0,(int)intFloor(radius*exptablescale))];
		      //cout << "Splat " << i << "s Weight to pixel (" << x << "," << y <<  ") is :: " << fragment.weight << endl;
		      fragment.z = z;
		      fragment.normal = n;
		      abuffer->appendFragment(x,y,0,fragment);
		    }
		}
	    }
	}
      
      if( pointvisualization )
	{
	  float w0 = 1.0f / p0h.w;
	  float midx = p0h.x * w0;
	  float midy = p0h.y * w0;
	  glPointSize(3.0f);
	  glBegin(GL_POINTS);
	  glVertex3f(midx*vpScale.x*figsx, midy*vpScale.y*figsy, -1);
	  glEnd();
	}
      
      if( reconstructionvisualization )
	{
	  /*		//draw bounding box
				glBegin(GL_LINE_LOOP);
				glVertex3f((xmin-vpMid.x)*figsx, (ymax-vpMid.y)*figsy, -1);
				glVertex3f((xmax-vpMid.x)*figsx, (ymax-vpMid.y)*figsy, -1);
				glVertex3f((xmax-vpMid.x)*figsx, (ymin-vpMid.y)*figsy, -1);
				glVertex3f((xmin-vpMid.x)*figsx, (ymin-vpMid.y)*figsy, -1);
				glEnd();
	  */
	  //draw splat boundary
	  float w0 = 1.0f / p0h.w;
	  float midx = p0h.x * w0;
	  float midy = p0h.y * w0;
	  
	  int segments = 25;
	  float a = 0.0f;
	  float aa = 2.0f * 3.1415f / float(segments);
	  Matrix2x2 J( Mi[0][0], Mi[0][1],
		       Mi[1][0], Mi[1][1] );
	  Matrix2x2 Jt = J;
	  Jt.transpose();
	  J *= Q;
	  J *= Jt;
	  float sk = sqrt(F);
	  glBegin(GL_LINE_LOOP);
	  for(int i=0;i<segments;++i)
	    {
	      Vector2 uv;
	      float f,g,t;
	      
	      //perspective correct position
	      // a*x^2+2*b*x*y+c*y^2 = k * (oow0 + dwx*x + dwy*y)^2
	      // sqrt(a*x^2+2*b*x*y+c*y^2) = sqrt(k) * (oow0 + dwx*x + dwy*y)
	      // y = f/g*x =>
	      // sqrt(a*x^2+2*b*x*f/g*x+c*f/g*x*f/g*x) = sqrt(k) * (oow0 + dwx*x + dwy*f/g*x)
	      // sqrt(a*x^2*g^2+2*b*x^2*f*g+c*f^2*x^2) = sqrt(k) * (oow0*g + dwx*x*g + dwy*f*x)
	      // sqrt(a*g^2+2*b*f*g+c*f^2) * x - sqrt(k)*dwx*g*x - sqrt(k)*dwy*f*x = sqrt(k)*oow0*g
	      // (sqrt(a*g^2+2*b*f*g+c*f^2) - sqrt(k)*dwx*g - sqrt(k)*dwy*f) * x = sqrt(k)*oow0*g
	      // x = sqrt(k)*oow0*g / (sqrt(a*g^2+2*b*f*g+c*f^2) - sqrt(k)*dwx*g - sqrt(k)*dwy*f)
	      // y = sqrt(k)*oow0*f / (sqrt(a*g^2+2*b*f*g+c*f^2) - sqrt(k)*dwx*g - sqrt(k)*dwy*f)
	      
	      f = sin(a);
	      g = cos(a);
	      t = sk * w0 / (sqrt(J[0][0] * g * g + 2.0f * J[0][1] * f * g + J[1][1] * f * f) - sk * (g * Mi[0][2] + f * Mi[1][2]));
	      glVertex3f(((midx + g * t)*vpScale.x)*figsx, ((midy + f * t)*vpScale.y)*figsy, -1);
	      
	      a += aa;
	    }
	  glEnd();
	}
    }
  //	svPrint("%f ",100.0f*(float)subpixelsPassed/(float)subpixelsTested);
}


//=======================================================================
//=======================================================================

void mainloop()
{
  cout << "Inside Main Loop" << endl;
  glClearColor( 0, 0, 0, 0 );
  glClear(GL_COLOR_BUFFER_BIT);
  glColor3f(1.0f,1.0f,1.0f);
  glEnable(GL_BLEND);
  glBlendFunc(GL_ONE,GL_ONE);

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  glScalef(defaultWidth/width,defaultHeight/height,0.01);
  gluPerspective(90,1,0.01,1000);
  //gluPerspective(90,1,-1,1000);
  glMatrixMode( GL_MODELVIEW );
  glPushMatrix();
  glLoadIdentity();
  
  float sx = pixelScale * (float)sizex / defaultWidth;
  //Change Here
  float sy = pixelScale * (float)sizey / defaultHeight;
  
  figsx = sx*2.0f/(float)sizex;
  figsy = -sy*2.0f/(float)sizey;
  
  //	static Timer t;
  //	t.sample();
  //	svPrint("fps: %f",1.0f/t.getDelta());
  
  static CameraControls cameraControls(1.0f,1.0f,0.f,false);//rotationScale,translationScale,inertia,framerateDependent);
  Matrix4x4 cameraToWorld = cameraControls.getCameraMatrix(kbdBuffer);
  Matrix4x4 worldToCamera;
  worldToCamera = cameraToWorld;
  worldToCamera.invert();

  static float a = 0.0f,b = 0.0f;
  if(kbdBuffer['3']) a -= 0.1f;
  if(kbdBuffer['4']) a += 0.1f;
  //	if(kbdBuffer['5']) b -= 0.1f;
  //	if(kbdBuffer['6']) b += 0.1f;
  Matrix4x4 objectToCamera;
  objectToCamera.ident();
  objectToCamera *= MatrixFactory::rotateY( -a );
  objectToCamera *= MatrixFactory::rotateX( -b );
  objectToCamera *= 3.0f;
  objectToCamera.setColumn(3,Vector4(0,0,10,1.0f));
  objectToCamera.transpose();
  worldToCamera.transpose();
  objectToCamera *= worldToCamera;
  
  const float	hfov = 90.0f;
  const float	vfov = 90.0f;
  const float front = 0.1f;
  const float	back = 1000.0f;
  
  float hx = 1.0f / float(tan(degreeToRad(hfov)*0.5f));
  float hy = 1.0f / float(tan(degreeToRad(vfov)*0.5f));
  float Q = back / (back - front);
#ifdef PROJECTION_PERSPECTIVE
  Matrix4x4 projection(
		       hx, 0,   0,        0,
		       0,  -hy, 0,        0,
		       0,  0,   Q,		   1,
		       0,  0,   -Q*front, 0);
#else
  Matrix4x4 projection(
		       0.2f, 0,     0,        0,
		       0,    -0.2f, 0,        0,
		       0,    0,     Q,        0,
		       0,    0,     -Q*front, 1);
#endif
  Vector4 viewport(0,0,(float)sizex,(float)sizey);
  
  static bool analytic = false;
  if(kbdBuffer['c'])
    {
      //		if( analytic ) svPrint("using the stochastic sampling pipeling\n");
      analytic = false;
    }
  if(kbdBuffer['v'])
    {
      //		if( !analytic ) svPrint("using the analytic filtering pipeling\n");
      analytic = true;
    }
  
  static bool zrange = true;
  if(kbdBuffer['h'])
    {
      //		if( zrange ) svPrint("using the z-threshold surface reconstruction\n");
      zrange = false;
    }
  if(kbdBuffer['j'])
    {
      //		if( !zrange ) svPrint("using the z-range surface reconstruction\n");
      zrange = true;
    }
  
  static float zthreshold = 0.1f;
  if(kbdBuffer['k'])
    {
      //zthreshold -= 0.025f;
      zthreshold -= 0.25f;
      cout<<"zthreshold : "<<zthreshold<<endl;
		//		svPrint("z-threshold = %f\n",zthreshold);
    }
  if(kbdBuffer['l'])
    {
      //zthreshold += 0.025f;
      zthreshold += 0.25f;
      cout<<"zthreshold : "<<zthreshold<<endl;
      //		svPrint("z-threshold = %f\n",zthreshold);
    }
  
  abuffer->clear();
  ///if( analytic )
  analyticPipeline(splats,abuffer,objectToCamera,projection,viewport,front,back);
  ///else
  ///stochasticPipeline(splats,abuffer,objectToCamera,projection,viewport,front,back);
  abuffer->collapseToFrameBuffer(buffer, Vector4(0.0f,0.0f,0.0f,1.0f),zrange,zthreshold);
  
  //render texture mapped quad
  glTexImage2D(GL_TEXTURE_2D, 0, 4, sizex, sizey, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, buffer);
  glBegin(GL_TRIANGLE_FAN);
  glTexCoord2f(0,0);
  glVertex3f(-sx,sy,-1);
  glTexCoord2f(1,0);
  glVertex3f(sx,sy,-1);
  glTexCoord2f(1,1);
  glVertex3f(sx,-sy,-1);
  glTexCoord2f(0,1);
  glVertex3f(-sx,-sy,-1);
  glTexCoord2f(0,0);
  glVertex3f(-sx,sy,-1);
  glEnd();
  
  glPopMatrix();
  glutSwapBuffers();
}

//=============================================================================
//=============================================================================

vector<Vector3> find_nearest_points(Vector3 vec, Cell* box) {
  
  float side = box->side;
  vector<Vector3> final;
  vector<Vector3> tmp;
  
  float xdiff = std::fabs((float) (box->center[0] - vec[0]));
  float ydiff = std::fabs((float) (box->center[1] - vec[1]));
  float zdiff = std::fabs((float) (box->center[2] - vec[2]));
  if(xdiff<=side/2 && ydiff<=side/2 && zdiff <= side/2) {
    if(box->isLeaf()) {
      for(vector<PatchPoint*>::iterator pp_itr=box->patchPoints.begin(); pp_itr!=box->patchPoints.end(); pp_itr++) {
	Vector3 p((*pp_itr)->coords[0],(*pp_itr)->coords[1],(*pp_itr)->coords[2]);
	final.push_back(p);
      }
      return final;
    }
    else {
      for(vector<Cell*>::iterator itr=box->children.begin();itr!=box->children.end();itr++) {
	tmp = find_nearest_points(vec,*itr);
	for(vector<Vector3>::iterator tmp_itr=tmp.begin();tmp_itr<tmp.end();tmp_itr++) {
	  final.push_back(*tmp_itr);
	}
      }
    }
  }
  return final;
}

vector<Vector3> find_nbrs(Vector3 pos) {
    
  vector<Vector3> tmp,final;
  vector<Vector3>::iterator tmp_itr;
  tmp = find_nearest_points(pos,root);
  for(tmp_itr=tmp.begin();tmp_itr<tmp.end();tmp_itr++) {
    final.push_back(*tmp_itr);
  }
  return final;
}

int surfel_render_main()
{
  cout << "Inside Sufel Render Main" << endl;
#ifdef OBJECT_MODEL

  getrusage(RUSAGE_SELF,&rstart);
  
  double maxl = .2;
  long int Vertices=0;

  for(PDListItr p_itr = surfelList.begin();p_itr != surfelList.end();p_itr++) {
    Vector3 coord((*p_itr)->coords[0],(*p_itr)->coords[1],-1*(*p_itr)->coords[2]);
    Vector3 normals((*p_itr)->normal[0],(*p_itr)->normal[1],-1*(*p_itr)->normal[2]);
    Vector4 radVal((*p_itr)->radVal[0],(*p_itr)->radVal[1],(*p_itr)->radVal[2],1.0);
    Splat st;
    st.m_position = coord;
    st.m_normal = normals;
    st.m_color = radVal;
    st.m_size = 0.0;
    splats.push_back(st);
  }
  
  float srtArry[NBD_SIZE];
  int count = 0;
 
#ifdef OPTIMIZE
  
  /*Finding Neighbors for Splat radius calculation Starts here.*/

  cout << "INSIDE THE OPTIMIZED CODE :)" << endl;

  for(int i=0;i<splats.size();i++) {
    double maxl = 0.0;
    for(int k=0;k<NBD_SIZE;k++) {
      srtArry[k] = MAX_SPLAT_SIZE;
    }
    vector<Vector3> tmp = find_nbrs(splats[i].m_position);
    
    for(int j=0;j<tmp.size();j++) {
      double l = (splats[i].m_position - tmp[j]).length();
      if(l<MAX_SPLAT_SIZE) {
	maxl = max2(l,maxl);
	//	cout<<l<<endl;
	int loc = 0;
	float tmax = -100.0;
	for(int m=0;m<NBD_SIZE;m++) {
	  tmax = max2(srtArry[m],tmax);
	  if(tmax<=srtArry[m]) {
	    loc = m;
	  }
	}
	if(srtArry[loc]>maxl) {
	  srtArry[loc] = maxl;
	}
      }
    }

    int loc = 0;
    float tmax = -100.0;
    for(int m=0;m<NBD_SIZE;m++) {
      tmax = max2(srtArry[m],tmax);
    }
    
    if(tmax<1e-4) {
      splats[i].m_size = MAX_SPLAT_SIZE;
    }
    else {
      splats[i].m_size = tmax;
    }
    //cout << "Splat[" << i << "] :: " << splats[i].m_size << endl;
    //tmp.clear();
  }
  
  /*Finding Neighbors for Splat radius calculation Ends here.*/

#else
  
  cout << "INSIDE THE UN-OPTIMIZED CODE :)" << endl;
  
  for(int i=0;i<splats.size();i++) {
    double maxl = 0.0;
    for(int k=0;k<NBD_SIZE;k++) {
      srtArry[k] = MAX_SPLAT_SIZE;
    }
    for(int j=0;j<splats.size();j++) {
      double l = (splats[i].m_position - splats[j].m_position).length();
      if(l<MAX_SPLAT_SIZE && (i!=j)) {
	maxl = max2(l,maxl);
	//	cout<<l<<endl;
	int loc = 0;
	float tmax = -100.0;
	for(int m=0;m<NBD_SIZE;m++) {
	  tmax = max2(srtArry[m],tmax);
	  if(tmax<=srtArry[m]) {
	    loc = m;
	  }
	}
	if(srtArry[loc]>=maxl) {
	  srtArry[loc] = maxl;
	}
      }
    }
    
    int loc = 0;
    float tmax = -100.0;
    for(int m=0;m<NBD_SIZE;m++) {
      tmax = max2(srtArry[m],tmax);
    }
    
    if(tmax<1e-4) {
      splats[i].m_size = MAX_SPLAT_SIZE;
    }
    else {
      splats[i].m_size = tmax;
    }
    //cout << "Splat[" << i << "] :: " << splats[i].m_size << endl;
  }
  
#endif
  
  getrusage(RUSAGE_SELF,&rend);
  cout << "Time Taken for Rendering :: " << rdiff(&rstart,&rend);

#elif defined OBJECT_CHESSBOARD
  int w = 128, h = 128;
  const float gscale = 0.05f;
  splats.resize(w*h);
  for(int y=0;y<h;++y)
    {
      for(int x=0;x<w;++x)
	{
	  int i = y*w+x;
	  splats[i].m_position.set(float(x-w/2)*gscale,-2.0f,-float(y-h/2)*gscale);
	  splats[i].m_normal.set(0,1,0);
	  splats[i].m_size = gscale * 1.0f;
	  splats[i].m_color = ((x>>4)^(y>>4))&1 ? Vector4(0,0,0,1) : Vector4(1,1,1,1);
	}
    }
#elif defined OBJECT_ONESPLAT
  splats.resize(1);
  splats[0].m_position.set(0,-1,0);
  splats[0].m_normal.set(0,1,0);
  splats[0].m_size = 1.2f;
  splats[0].m_color.set(192,128,0,1);
#endif
  
  abuffer = new ABuffer(sizex,sizey,subpixels);
  buffer = new Color[sizex * sizey];
  makeJitter(subpixels);
  
  memset(kbdBuffer,0,256*sizeof(char));
  
  int qwer = 1;
  char* empty = "test.exe";
  glutInit( &qwer, &empty );
  
  glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB );
  glutInitWindowSize( (int)defaultWidth, (int)defaultHeight );
  glutCreateWindow( "Super 3D Virtual Reality" );
  
  glutDisplayFunc( mainloop );

  glutSetKeyRepeat(GLUT_KEY_REPEAT_ON);
  glutKeyboardFunc( kbdfunc );
  glutSpecialFunc( specialfunc );
  glutKeyboardUpFunc( kbdupfunc );
  glutSpecialUpFunc( specialupfunc );
  glutReshapeFunc( reshape );
  glutIdleFunc( idlefunc );
  
  glEnable(GL_TEXTURE_2D);
  
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
  glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_DECAL);
  
  glutMainLoop();
  return 0;
}

//=======================================================================
//=======================================================================
