#include<math.h>
#include <GL/gl.h>
#include <GL/glut.h>

#include<iostream>
#include "Radical.hpp"
#include<vector>
using namespace std;


#define BASE_Z 28
#define BASE_ANG_X 0.0
#define BASE_ANG_Y 0.0

GLfloat eye_z = BASE_Z;
GLfloat rot_x = BASE_ANG_X;
GLfloat rot_y = BASE_ANG_Y;
GLfloat rot_z = 0.0;

class PointData {
public:

  PointData(Vec3f coords,Vec3f normal,Vec3f radVal,float radius)
    :coords(coords),normal(normal),radVal(radVal), radius(radius){}

  float radius;
  Vec3f coords;
  Vec3f normal;
  Vec3f radVal;
};
typedef vector<PointData*>           PDList;
typedef vector<PointData*>::iterator PDListItr;

PDList surfelList;

long int readSurfels(vector<PointData*>&, char*);
void normalize(vector<PointData*>&, long int);

long int numVertices=0;



int screenWidth = 800;     //Height and width of the screen window
int screenHeight = 600;

//Height and width of world window
float windowLeft, windowRight, windowNear, windowFar, windowBottom, windowTop;

float windowWidth = windowRight-windowLeft;
float windowHeight = windowTop-windowBottom;

const float center_wW = windowWidth / 2.0;   //Center of world window
const float center_wH = windowHeight / 2.0;

GLdouble AR = windowWidth / windowHeight;    //Aspect ratio of world window


/*****************Function for setting the world window *******************/

void setWindow(float left, float right, float bottom, float top, float near, float far)       
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(45,1.0,0.0,100.0);
//    gluPerspective(45,1.0, 0, 100);
//    glMatrixMode(GL_MODELVIEW);
//    glLoadIdentity();
//    gluLookAt(0.0,0.0,30.0,0.0,0.0,0.0,0.0,1.0,0.0);
}

/***************** Function for setting the viewport*******************/

void setViewport(int left, int right, int bottom, int top)
{
    glViewport(left, bottom, right - left, top - bottom);
}

/***************** Initialization of the environment*******************/

void myInit(void)
{
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glPointSize(5.0);
    setWindow(windowLeft, windowRight, windowBottom, windowTop,windowNear, windowNear-windowFar);
    setViewport(0, screenWidth, 0, screenHeight);
}


/*****************Draws the point on the screen with specified color *******************/

void drawDot(GLfloat x, GLfloat y, GLfloat z, GLfloat r, GLfloat g, GLfloat b)
{
//	cout<<"Red : "<<r<<"Green: "<<g<<"Blue: "<<b<<endl;
	if(r<0) r=0; //-1*r;
	if(g<0) g=0; //g*-1;
	if(b<0) b=0; //b*-1;
//	glColor3d(0,0,0);
//	glBegin(GL_LINE);
//                glVertex3f(0,0,0);
//		glVertex3f(1,1,1);
//        glEnd();
	glColor3f(r,g,b);
	glBegin(GL_POINTS);
		glVertex3f(x,y,z);
	glEnd();

}

/*****************Keyboard function  *******************/

void myKeyboard(unsigned char key, GLint x, GLint y) {
  GLfloat rval = 5.0;
  GLfloat val = .1;
  switch(key) {
  case 's':
    eye_z += val;
    break;
  case 'S':
    eye_z -= val;
    break;
  case 'x':
    rot_x += rval;
    break;
  case 'y':
    rot_y += rval;
    break;
  case 'z':
    rot_z += rval;
     break;
  case 'X':
    rot_x -= rval;
    break;
  case 'Y':
    rot_y -= rval;
    break;
  case 'Z':
    rot_z -= rval;
    break;
  case 'q':
    rot_x = BASE_ANG_X;
    rot_y = BASE_ANG_Y;
    rot_z = 0;
    eye_z = BASE_Z;
    break;
  default:
    break;
   }
  glutPostRedisplay();
}


/***************** Function to control the display *******************/

void myDisplay(void)
{
    glClearColor(0.0, 0.0, 0.0, 0.0); 
    glClearDepth(0.0); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(0.0,0.5,eye_z+5,0.0,0.5,0.0,0.0,1.0,0.0);
    glRotatef(rot_x,1,0,0);
    glRotatef(rot_y,0,1,0);
    glRotatef(rot_z,0,0,1);
//    glViewport(0,0,800,600);

    for(PDListItr p_itr = surfelList.begin();p_itr != surfelList.end();p_itr++) 
    {
      drawDot((*p_itr)->coords[0],(*p_itr)->coords[1],(*p_itr)->coords[2],(*p_itr)->radVal[0],(*p_itr)->radVal[1],(*p_itr)->radVal[2]);
    }
    glutSwapBuffers();
    glFlush();
}


/***************** Function to control the shape of the screen window*******************/

//Maintains the SAME ASPECT RATIO even if the screen is resized

void myReshape(GLsizei W, GLsizei H)
{
    int WiS, WiE, HeS, HeE;

    if (AR > W / H) {
//	HeS = (int) (H / 2 - W / (2 * AR));
//	HeE = (int) (H / 2 + W / (2 * AR));
//	setViewport(0, W, HeS, HeE);
	setViewport(0, W, 0, int(W/AR));
    } else {
//	WiS = (int) (W / 2 - (H * AR) / 2);
//	WiE = (int) (W / 2 + (H * AR) / 2);
//	setViewport(WiS, WiE, 0, H);
	setViewport(0, int(H*AR), 0, H);
    }

     glutPostRedisplay();
}


/*****************Normalizing the color values and finding window width and height*******************/

void normalize(vector<PointData*>& pts,long int Vertices)
{
	Vec3f* colors=new Vec3f[Vertices];
	Vec3f* points=new Vec3f[Vertices];
	Vec3f pMax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL), pMin(HUGE_VAL,HUGE_VAL,HUGE_VAL);
	Vec3f cMax(-HUGE_VAL,-HUGE_VAL,-HUGE_VAL);
	unsigned long int i=0;
	for(PDListItr p_itr = pts.begin();p_itr!=pts.end();p_itr++){
	    for(int j=0;j<3;j++){
	      colors[i][j] = (*p_itr)->radVal[j];  
	      cMax[j]=max(colors[i][j],cMax[j]);
	      points[i][j] = (*p_itr)->coords[j];
	      pMax[j]=max(points[i][j],pMax[j]);
	      pMin[j]=min(points[i][j],pMin[j]);
	    }
	    i++;
	  }

	for(PDListItr p_itr = pts.begin();p_itr!=pts.end();p_itr++)
	    for(int j=0;j<3;j++)
		(*p_itr)->radVal[j] = (*p_itr)->radVal[j]/cMax[j]*50;

	windowLeft  = pMin[0];
	windowRight = pMax[0];
	windowBottom = pMin[1];
	windowTop = pMax[1];
	windowNear = pMax[2];
	windowFar = pMin[2];

	cout<<"windowLeft is:"<<windowLeft<<endl;
	cout<<"windowRight is:"<<windowRight<<endl;
	cout<<"windowBottom is:"<<windowBottom<<endl;
	cout<<"windowTop is:"<<windowTop<<endl;
	cout<<"windowNear is:"<<windowNear<<endl;
	cout<<"windowFar is:"<<windowFar<<endl;

}


/*****************Reads the surfel data from a file*******************/


long int readSurfels(vector<PointData*>& pts,char* filename){
FILE *fp_pointD = fopen("surfelpointCoord","w");
long int Vertices=0;
float a,b,c;
ifstream in;
in.open(filename);

  while(in)
  {
        in>>a>>b>>c;
        Vec3f coord(a,b,c);
//	cout<<"X coord is "<<a<<"\t Y coord is "<<b<<"\t Z coord is "<<c<<endl;
        fprintf(fp_pointD,"%f \t %f \t %f\n",a, b, c);	
        in>>a>>b>>c;
        Vec3f normals(a,b,c);
        in>>a>>b>>c;
        Vec3f radVal(a,b,c);
        in>>a;
        float radius = a;
  Vertices++;
  PointData* p=new PointData(coord,normals,radVal,radius);
  pts.push_back(p);

  }
  in.close();
  fclose(fp_pointD);  
  if(pts.size() > 0)
     pts.pop_back();
 return(Vertices);
}

/*****************The main program*******************/

int main(int argc, char **argv)
{
    numVertices = readSurfels(surfelList, argv[1]);
    normalize(surfelList, numVertices);
    glutInit(&argc, argv);	// initialize the toolkit
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);	// set display mode
    glutInitWindowSize(screenWidth, screenHeight);	// set window size
    glutInitWindowPosition(100, 150);	// set window position on screen
    glutCreateWindow("my first attempt");	// open the screen window
    glutDisplayFunc(myDisplay);	// register redraw function
    glutKeyboardFunc(myKeyboard);
    glutReshapeFunc(myReshape);
    
    myInit();
    glutMainLoop();		// go into a perpetual loop
    return 0;
}
