#include <stdio.h>

#include <math.h>

#include <ctype.h>

#include <string.h>

#include "../mgflib/parser.h"

#include "../mgflib/lookup.h"

#include "../mgflib/xf.c"


#define MAXID		48	/* maximum identifier length */

#define VERTFMT		"%+16.9e %+16.9e %+16.9e\n%+6.3f %+6.3f %+6.3f"
#define PRINT		"[(%6.6f %6.6f %6.6f),(%6.6f %6.6f %6.6f)]"
#define VZVECT		"+0.000 +0.000 +0.000"
#define VFSEPPOS	50	/* position of newline in above */
#define VFLEN		72	/* total vertex string length */
#define MAXVERT		10240	/* maximum cached vertices */

#define setvkey(k,v)	sprintf(k,VERTFMT,(v)->p[0],(v)->p[1],(v)->p[2],\
					(v)->n[0],(v)->n[1],(v)->n[2]);

char	vlist[MAXVERT][VFLEN];	/* our vertex cache */
int	nverts;			/* current cache size */

LUTAB	vert_tab = LU_SINIT(NULL,NULL);

struct face {
	struct face	*next;		/* next face in list */
	short		nv;		/* number of vertices */
	short		vl[3];		/* vertex index list (variable) */
}	*flist, *flast;		/* our face cache */

#define newface(n)	(struct face *)malloc(sizeof(struct face) + \
				((n) > 3 ? (n)-3 : 0)*sizeof(short))
#define freeface(f)	free((MEM_PTR)f)

#define	TABSTOP		8	/* assumed number of characters per tab */
#define	SHIFTW		2	/* nesting shift width */
#define MAXIND		15	/* maximum indent level */

char	tabs[MAXIND*SHIFTW+1];	/* current tab-in string */

#define curmatname	(c_cmname == NULL ? "mat" : to_id(c_cmname))
 
//extern char *stmt[500][50];
//extern int flag;
extern int	i_comment(), i_object(), i_xf(),
		i_cyl(), i_face(), i_sph();

extern char	*to_id();

typedef struct {
	int color;
        float x;
        float y;
        float z;	
//	float nx,ny,nz;	
} vertex;


/* A triangle made up of three vertices */



#define V_MAX 100       /* Maximum number of vertices allowed (arbitrary) */

#define BIG 1.0e30      /* A number bigger than we expect to find here */

#define COUNTER_CLOCKWISE 0
#define CLOCKWISE 1

float	rgbvalD[3], rgbvalE[3];


/*
 * orientation
 *
 * Return either clockwise or counter_clockwise for the orientation
 * of the polygon.
 */

int orientation( n, v )
    int n;                      /* Number of vertices */
    vertex v[];                 /* The vertex list */
{
    float area;
    int i;

    /* Do the wrap-around first */
    area = v[n-1].x * v[0].y - v[0].x * v[n-1].y;

    /* Compute the area (times 2) of the polygon */
    for (i = 0; i < n-1; i++)
        area += v[i].x * v[i+1].y - v[i+1].x * v[i].y;

    if (area >= 0.0)
        return COUNTER_CLOCKWISE;
    else
        return CLOCKWISE;
} /* End of orientation */



/*
 * determinant
 *
 * Computes the determinant of the three points.
 * Returns whether the triangle is clockwise or counter-clockwise.
 */

int determinant( p1, p2, p3, v )
    int p1, p2, p3;             /* The vertices to consider */
    vertex v[];                 /* The vertex list */
{
    float x1, x2, x3, y1, y2, y3;
    float determ;

    x1 = v[p1].x;
    y1 = v[p1].y;
    x2 = v[p2].x;
    y2 = v[p2].y;
    x3 = v[p3].x;
    y3 = v[p3].y;

    determ = (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1);
    if (determ >= 0.0)
        return COUNTER_CLOCKWISE;
    else
        return CLOCKWISE;
} /* End of determinant */



/*
 * distance2
 *
 * Returns the square of the distance between the two points
 */

float distance2( x1, y1, x2, y2 )
    float x1, y1, x2, y2;
{
    float xd, yd;               /* The distances in X and Y */
    float dist2;                /* The square of the actual distance */

    xd = x1 - x2;
    yd = y1 - y2;
    dist2 = xd * xd + yd * yd;

    return dist2;
} /* End of distance2 */



/*
 * no_interior
 *
 * Returns 1 if no other point in the vertex list is inside
 * the triangle specified by the three points.  Returns
 * 0 otherwise.
 */

int no_interior( p1, p2, p3, v, vp, n, poly_or )
    int p1, p2, p3;             /* The vertices to consider */
    vertex v[];                 /* The vertex list */
    int vp[];                   /* The vertex pointers (which are left) */
    int n;                      /* Number of vertices */
    int poly_or;                /* Polygon orientation */
{
    int i;                      /* Iterative counter */
    int p;                      /* The test point */
    for (i = 0; i < n; i++) {
        p = vp[i];              /* The point to test */
        if ((p == p1) || (p == p2) || (p == p3))
            continue;           /* Don't bother checking against yourself */
        if (   (determinant( p2, p1, p, v ) == poly_or)
            || (determinant( p1, p3, p, v ) == poly_or)
            || (determinant( p3, p2, p, v ) == poly_or) ) {
            continue;           /* This point is outside */
        } else {
            return 0;           /* The point is inside */
        }
    }
    return 1;                   /* No points inside this triangle */
} /* End of no_interior */



/*
 * draw_poly
 *
 * Call this procedure with a polygon, this divides it into triangles
 * and calls the triangle routine once for each triangle.
 *
 * Note that this does not work for polygons with holes or self
 * penetrations.
 */
draw_triangle( v1,v2,v3 )
vertex v1,v2,v3;
{

        printf("%.3f %.3f %.3f ",v1.x,v1.y,v1.z);
	printf("%.3f %.3f %.3f ",v2.x,v2.y,v2.z);
	printf("%.3f %.3f %.3f ",v3.x,v3.y,v3.z);

	printf("%.3f %.3f %.3f ",rgbvalD[0], rgbvalD[1], rgbvalD[2]);
	printf("%.3f %.3f %.3f \n",rgbvalE[0], rgbvalE[1], rgbvalE[2]);

}
void
mult4(m4a, m4b, m4c)         /* multiply m4b X m4c and put into m4a */
register float  m4a[1][4],m4b[1][4];
register MAT4 m4c;
{
        register int  i, j;
	float m4tmp[1][4];
        for (i = 0 ; i<4 ; i++ )
                 m4tmp[0][i]  =  m4b[0][0]*m4c[0][i] +
                                 m4b[0][1]*m4c[1][i] +
                                 m4b[0][2]*m4c[2][i] +
                                 m4b[0][3]*m4c[3][i];

	 for (i = 0 ; i<4;i++ ){
		m4a[0][i]=m4tmp[0][i];
//		printf("%6.6g",m4a[0][i]);
	}
}

apply_transform(v,n)
vertex v[];
int n;
{
	float v1[1][4],res[1][4];
	int i;
	register XF_SPEC *sp;
	for(i=0;i<n;i++) {
	        v1[0][0]=v[i].x;
        	v1[0][1]=v[i].y;
	        v1[0][2]=v[i].z;
        	v1[0][3]=1;
		sp=xf_context;
		while(sp!=NULL){
		        mult4(v1,v1,sp->xf.xfm);
			sp=sp->prev;
		}
		v[i].x=v1[0][0];
		v[i].y=v1[0][1];
		v[i].z=v1[0][2];
		
//        	printf("%d %d\n",i,n);
	}
}



draw_poly( n, v )
    int n;                      /* Number of vertices in triangle */
    vertex v[];                 /* The vertex list (implicit closure) */
{
    int prev, cur, next;        /* Three points currently being considered */
    int vp[V_MAX];              /* Pointers to vertices still left */
    int count;                  /* How many vertices left */
    int min_vert;               /* Vertex with minimum distance */
    int i;                      /* Iterative counter */
    float dist;                 /* Distance across this one */
    float min_dist;             /* Minimum distance so far */
    int poly_orientation;       /* Polygon orientation */
    vertex v1,v2,v3;                 /* Triangle structure */


    apply_transform(v,n);

    if (n > V_MAX) {
        fprintf( stderr, "Error, more than %d vertices.\n", V_MAX);
        return;
    }

//	for (i=0;i<n;i++)
//	{	printf("%f %f %f,",v[i].x,v[i].y,v[i].z);
//	}printf("\n");
    poly_orientation = orientation( n, v );

    for (i = 0; i <= n; i++)
        vp[i] = i;              /* Put vertices in order to begin */

/* Slice off clean triangles until nothing remains */

    count = n;

    while (count > 3) {

        min_dist = BIG;         /* A real big number */
        min_vert = 0;           /* Just in case we don't find one... */
        for (cur = 0; cur < count; cur++) {
            prev = cur - 1;
            next = cur + 1;

            if (cur == 0)       /* Wrap around on the ends */
                prev = count - 1;
            else if (cur == count)
                next = 0;
            /* Pick out shortest distance that forms a good triangle */
            if (   (determinant( vp[prev], vp[cur], vp[next], v ) == poly_orientation)
                    /* Same orientation as polygon */
                && no_interior( vp[prev], vp[cur], vp[next], v, vp, count, poly_orientation )
                    /* No points inside */
                && ((dist = distance2( v[vp[prev]].x, v[vp[prev]].y,
                                       v[vp[next]].x, v[vp[next]].y )) < min_dist) )
                    /* Better than any so far */
            {
                min_dist = dist;
                min_vert = cur;
            }
        } /* End of for each vertex (cur) */

        /* The following error should "never happen". */
        if (min_dist == BIG)
            fprintf( stderr, "Error: Didn't find a triangle.\n" );

        prev = min_vert - 1;
        next = min_vert + 1;
        if (min_vert == 0)      /* Wrap around on the ends */
            prev = count - 1;
        else if (min_vert == (count-1))
            next = 0;

/* Output this triangle */

        v1.x = v[vp[prev]].x;
        v1.y = v[vp[prev]].y;
        v1.z = v[vp[prev]].z;
        v1.color = v[vp[prev]].color;
        v2.x = v[vp[min_vert]].x;
        v2.y = v[vp[min_vert]].y;
        v2.z = v[vp[min_vert]].z;
        v2.color = v[vp[min_vert]].color;
        v3.x = v[vp[next]].x;
        v3.y = v[vp[next]].y;
        v3.z = v[vp[next]].z;
        v3.color = v[vp[next]].color;

        draw_triangle( v1,v2,v3 );

/* Remove the triangle from the polygon */

        count -= 1;
        for (i = min_vert; i < count; i++)
            vp[i] = vp[i+1];
    }

/* Output the final triangle */

        v1.x = v[vp[0]].x;
        v1.y = v[vp[0]].y;
        v1.z = v[vp[0]].z;
        v1.color = v[vp[0]].color;
        v2.x = v[vp[1]].x;
        v2.y = v[vp[1]].y;
        v2.z = v[vp[1]].z;
        v2.color = v[vp[1]].color;
        v3.x = v[vp[2]].x;
        v3.y = v[vp[2]].y;
        v3.z = v[vp[2]].z;
        v3.color = v[vp[2]].color;
        draw_triangle( v1,v2,v3 );
} 


main(argc, argv)
int	argc;
char	*argv[];
{
	int	i;
				/* initialize dispatch table */
//	mg_ehand[MG_E_COMMENT] = i_comment;	/* we pass comments */
	mg_ehand[MG_E_COLOR] = c_hcolor;	/* they get color */
	mg_ehand[MG_E_CMIX] = c_hcolor;		/* they mix colors */
	mg_ehand[MG_E_CSPEC] = c_hcolor;	/* they get spectra */
	mg_ehand[MG_E_CXY] = c_hcolor;		/* they get chromaticities */
	mg_ehand[MG_E_CCT] = c_hcolor;		/* they get color temp's */
//	mg_ehand[MG_E_CYL] = i_cyl;		/* we do cylinders */
	mg_ehand[MG_E_ED] = c_hmaterial;	/* they get emission */
	mg_ehand[MG_E_FACE] = i_face;		/* we do faces */
	mg_ehand[MG_E_MATERIAL] = c_hmaterial;	/* they get materials */
	mg_ehand[MG_E_NORMAL] = c_hvertex;	/* they get normals */
	mg_ehand[MG_E_OBJECT] = i_object;	/* we track object names */
	mg_ehand[MG_E_POINT] = c_hvertex;	/* they get points */
	mg_ehand[MG_E_RD] = c_hmaterial;	/* they get diffuse refl. */
	mg_ehand[MG_E_RS] = c_hmaterial;	/* they get specular refl. */
	mg_ehand[MG_E_SIDES] = c_hmaterial;	/* they get # sides */
//	mg_ehand[MG_E_SPH] = i_sph;		/* we do spheres */
	mg_ehand[MG_E_TD] = c_hmaterial;	/* they get diffuse trans. */
	mg_ehand[MG_E_TS] = c_hmaterial;	/* they get specular trans. */
	mg_ehand[MG_E_VERTEX] = c_hvertex;	/* they get vertices */
	mg_ehand[MG_E_XF] = i_xf;		/* we track transforms */
//	mg_ehand[MG_E_XF] = xf_handler;		/* we track transforms */
	mg_init();		/* initialize the parser */

	for (i=1 ; i < argc; i++) {
//		printf("## %s %s ##############################\n",
//				argv[0], argv[i]);
		mg_nunknown = 0;
		if (mg_load(argv[i]) != MG_OK)
			exit(1);
//		if (mg_nunknown)
//			printf("## %s %s: %u unknown entities\n",
//					argv[0], argv[i], mg_nunknown);
	}
	flush_cache();		/* flush face cache, just in case */
//	printf("}\n");		/* close root node */
	exit(0);
}


indent(deeper)				/* indent in or out */
int	deeper;
{
	static int	level;		/* current nesting level */
	register int	i;
	register char	*cp;

	if (deeper) level++;		/* in or out? */
	else if (level > 0) level--;
					/* compute actual shift */
	if ((i = level) > MAXIND) i = MAXIND;
	cp = tabs;
	for (i *= SHIFTW; i >= TABSTOP; i -= TABSTOP)
		*cp++ = '\t';
	while (i--)
		*cp++ = ' ';
	*cp = '\0';
}


int
i_comment(ac, av)			/* transfer comment as is */
int	ac;
char	**av;
{
	fputs(tabs, stdout);
	putchar('#');			/* Inventor comment character */
	while (--ac > 0) {
		putchar(' ');
		fputs(*++av, stdout);
	}
	putchar('\n');
	return(MG_OK);
}


int
i_object(ac, av)			/* group object name */
int	ac;
char	**av;
{
	static int	objnest;

	flush_cache();			/* flush cached objects */
	if (ac == 2) {				/* start group */
//		printf("%sDEF %s Group {\n", tabs, to_id(av[1]));
//		indent(1);
		objnest++;
		return(MG_OK);
	}
	if (ac == 1) {				/* end group */
		if (--objnest < 0)
			return(MG_ECNTXT);
//		indent(0);
//		fputs(tabs, stdout);
//		fputs("}\n", stdout);
		return(MG_OK);
	}
	return(MG_EARGC);
}


int
i_xf(ac, av)				/* transform object(s) */
int	ac;
char	**av;
{
	static long	xfid;
	register XF_SPEC	*spec;

	flush_cache();			/* flush cached objects */
	if (ac == 1) {			/* end of transform */
		if ((spec = xf_context) == NULL)
			return(MG_ECNTXT);
//		indent(0);			/* close original segment */
//		printf("%s}\n", tabs);
//		indent(0);
//		printf("%s}\n", tabs);
		if (spec->xarr != NULL) {	/* check for iteration */
			register struct xf_array	*ap = spec->xarr;
			register int	n;

			ap->aarg[ap->ndim-1].i = 1;	/* iterate array */
			for ( ; ; ) {
				n = ap->ndim-1;
				while (ap->aarg[n].i < ap->aarg[n].n) {
					sprintf(ap->aarg[n].arg, "%d",
							ap->aarg[n].i);
//					printf("%sSeparator {\n", tabs);
//					indent(1);
					(void)put_xform(spec);
//					printf("%sUSE _xf%ld\n", tabs,
//							spec->xid);
//					indent(0);
//					printf("%s}\n", tabs);
					++ap->aarg[n].i;
				}
				ap->aarg[n].i = 0;
				(void)strcpy(ap->aarg[n].arg, "0");
				while (n-- && ++ap->aarg[n].i >= ap->aarg[n].n) {
					ap->aarg[n].i = 0;
					(void)strcpy(ap->aarg[n].arg, "0");
				}
				if (n < 0)
					break;
				sprintf(ap->aarg[n].arg, "%d", ap->aarg[n].i);
			}
		}
						/* pop transform */
		xf_context = spec->prev;
		free_xf(spec);
		return(MG_OK);
	}
					/* else allocate new transform */
	if ((spec = new_xf(ac-1, av+1)) == NULL)
		return(MG_EMEM);
	spec->xid = ++xfid;		/* assign unique ID */
	spec->prev = xf_context;	/* push onto stack */
	xf_context = spec;
					/* translate xf specification */
//	printf("%sSeparator {\n", tabs);
//	indent(1);
	if (put_xform(spec) < 0)
		return(MG_ETYPE);
//	printf("%sDEF _xf%ld Group {\n", tabs, spec->xid);	/* begin */
//	indent(1);
	return(MG_OK);
}


int
put_xform(spec)			/* translate and print transform */
register XF_SPEC	*spec;
{
	register char	**av;
	register int	n;

	n = xf_ac(spec) - xf_ac(spec->prev);
	if (xf(&spec->xf, n, av=xf_av(spec)) != n)
		return(-1);
//	printf("%sMatrixTransform {\n", tabs);
//	indent(1);
//	printf("%s# xf", tabs);		/* put out original as comment */
//	while (n--) {
//		putchar(' ');
//		fputs(*av++, stdout);
//	}
//	putchar('\n');			/* put out computed matrix */
//	printf("%smatrix %13.9g %13.9g %13.9g %13.9g\n", tabs,
//		spec->xf.xfm[0][0], spec->xf.xfm[0][1],
//			spec->xf.xfm[0][2], spec->xf.xfm[0][3]);
//	for (n = 1; n < 4; n++)
//		printf("%s       %13.9g %13.9g %13.9g %13.9g\n", tabs,
//				spec->xf.xfm[n][0], spec->xf.xfm[n][1],
//				spec->xf.xfm[n][2], spec->xf.xfm[n][3]);
//	indent(0);
//	printf("%s}\n", tabs);
	return(0);
}


int
put_material()			/* put out current material */
{
	char	*mname = curmatname;
//	float	rgbval[3];

	if (!c_cmaterial->clock) {	/* current, just use it */

		mgf2rgb(&c_cmaterial->rd_c, c_cmaterial->rd, rgbvalD);
//		printf("%sdiffuseColor %.4f %.4f %.4f\n", tabs,	rgbvalD[0], rgbvalD[1], rgbvalD[2]);
		if (c_cmaterial->ed > FTINY) {
			mgf2rgb(&c_cmaterial->ed_c, 1.0, rgbvalE);
//			printf("%semissiveColor %.4f %.4f %.4f\n",tabs,rgbvalE[0], rgbvalE[1], rgbvalE[2]);
		}

//		printf("%sUSE %s\n", tabs, mname);
		return(0);
	}
				/* else update definition */
//	printf("%sDEF %s Group {\n", tabs, mname);
//	indent(1);
//	printf("%sMaterial {\n", tabs);
//	indent(1);
	mgf2rgb(&c_cmaterial->rd_c, c_cmaterial->rd, rgbvalD);
/*	printf("%sambientColor %.4f %.4f %.4f\n", tabs,
			rgbval[0], rgbval[1], rgbval[2]);*/
//	printf("%sdiffuseColor %.4f %.4f %.4f\n", tabs,
//			rgbvalD[0], rgbvalD[1], rgbvalD[2]);
/*	if (c_cmaterial->rs > FTINY) {
		mgf2rgb(&c_cmaterial->rs_c, c_cmaterial->rs, rgbval);
		printf("%sspecularColor %.4f %.4f %.4f\n", tabs,
				rgbval[0], rgbval[1], rgbval[2]);
		printf("%sshininess %.3f\n", tabs, 1.-sqrt(c_cmaterial->rs_a));
	}*/
	if (c_cmaterial->ed > FTINY) {
		mgf2rgb(&c_cmaterial->ed_c, 1.0, rgbvalE);
//		printf("%semissiveColor %.4f %.4f %.4f\n", tabs,
//				rgbvalE[0], rgbvalE[1], rgbvalE[2]);
	}
	else{
		rgbvalE[0]=0.0;
		rgbvalE[1]=0.0;	
		rgbvalE[2]=0.0;
	}
/*	if (c_cmaterial->ts > FTINY)
		printf("%stransparency %.4f\n", tabs,
				c_cmaterial->ts + c_cmaterial->td);*/
//	indent(0);
//	printf("%s}\n", tabs);
//	indent(0);
//	printf("%s}\n", tabs);
	c_cmaterial->clock = 0;
	return(0);
}


int
i_face(ac, av)			/* translate an N-sided face */
int	ac;
char	**av;
{
	static char	lastmat[MAXID];
	struct face	*newf;
	register C_VERTEX	*vp;
	register LUENT	*lp;
	register int	i;

	if (ac < 4)
		return(MG_EARGC);
	if ( strcmp(lastmat, curmatname) || c_cmaterial->clock ||
			nverts == 0 || nverts+ac-1 >= MAXVERT) {
		flush_cache();			/* new cache */
		lu_init(&vert_tab, MAXVERT);
		//printf("%sSeparator {\n", tabs);
		indent(1);
		if (put_material() < 0)		/* put out material */
			return(MG_EBADMAT);
		(void)strcpy(lastmat, curmatname);
	}
				/* allocate new face */
	if ((newf = newface(ac-1)) == NULL)
		return(MG_EMEM);
	newf->nv = ac-1;
				/* get vertex references */
	vertex v[newf->nv];
	
	for (i = 0; i < newf->nv; i++) {
		if ((vp = c_getvert(av[i+1])) == NULL)
			return(MG_EUNDEF);
		setvkey(vlist[nverts], vp);

		v[i].x=vp->p[0];
		v[i].y=vp->p[1];
		v[i].z=vp->p[2];
//		v[i].nx=(vp->n[0]<FTINY)?vp->n[0]:0.0;
//		v[i].ny=(vp->n[1]<FTINY)?vp->n[1]:0.0;
//		v[i].nz=(vp->n[2]<FTINY)?vp->n[2]:0.0;

		lp = lu_find(&vert_tab, vlist[nverts]);
		if (lp == NULL)
			return(MG_EMEM);
		if (lp->key == NULL)
			lp->key = (char *)vlist[nverts++];
		newf->vl[i] = ((char (*)[VFLEN])lp->key - vlist);
	}

	draw_poly((newf->nv),v);
				/* add to face list */
	newf->next = NULL;
	if (flist == NULL)
		flist = newf;
	else
		flast->next = newf;
	flast = newf;
	return(MG_OK);		/* we'll actually put it out later */
}

char *
to_id(name)			/* make sure a name is a valid Inventor ID */
register char	*name;
{
	static char	id[MAXID];
	register char	*cp;

	for (cp = id; *name && cp < MAXID-1+id; name++)
		if (isalnum(*name) || *name == '_')
			*cp++ = *name;
		else
			*cp++ = '_';
	*cp = '\0';
	return(id);
}


flush_cache()			/* put out cached faces */
{
	int	donorms = 0;
	register struct face	*f;
	register int	i;

	if (nverts == 0)
		return;
					/* put out coordinates */
//	printf("%sCoordinate3 {\n", tabs);
//	indent(1);
	vlist[0][VFSEPPOS] = '\0';
//	printf("%spoint [ %s", tabs, vlist[0]);
//	for (i = 1; i < nverts; i++) {
//		vlist[i][VFSEPPOS] = '\0';
//		printf(",\n%s        %s", tabs, vlist[i]);
//		if (strcmp(VFSEPPOS+1+vlist[i], VZVECT))
//			donorms++;
//	}
//	indent(0);
//	printf(" ]\n%s}\n", tabs);
//	if (donorms) {			/* put out normals */
//		printf("%sNormal {\n", tabs);
//		indent(1);
//		printf("%svector [ %s", tabs, VFSEPPOS+1+vlist[0]);
//		for (i = 1; i < nverts; i++)
//			printf(",\n%s         %s", tabs, VFSEPPOS+1+vlist[i]);
//		indent(0);
//		printf(" ]\n%s}\n", tabs);
//	}*/
					/* put out faces */
//	printf("%sIndexedFaceSet {\n", tabs);
//	indent(1);
	f = flist;			/* coordinate indices */
//	printf("%scoordIndex [ %d", tabs, f->vl[0]);
//	for (i = 1; i < f->nv; i++)
//		printf(", %d", f->vl[i]);
//	for (f = f->next; f != NULL; f = f->next) {
//		printf(", -1,\n%s             %d", tabs, f->vl[0]);
//		for (i = 1; i < f->nv; i++)
//			printf(", %d", f->vl[i]);
//	}
//	printf(" ]\n");
//	if (donorms) {
//		f = flist;			/* normal indices */
//		printf("%snormalIndex [ %d", tabs, f->vl[0]);
//		for (i = 1; i < f->nv; i++)
//			printf(", %d", f->vl[i]);
//		for (f = f->next; f != NULL; f = f->next) {
//			printf(", -1,\n%s              %d", tabs, f->vl[0]);
//			for (i = 1; i < f->nv; i++)
//				printf(", %d", f->vl[i]);
//		}
//		printf(" ]\n");
//	}
//	indent(0);			/* close IndexedFaceSet */
//	printf("%s}\n", tabs);
//	indent(0);			/* close face group */
//	printf("%s}\n", tabs);
	while ((f = flist) != NULL) {	/* free face list */
		flist = f->next;
		freeface(f);
	}
	lu_done(&vert_tab);		/* clear lookup table */
	nverts = 0;
}


