#include <math.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>

#include "vl/VLf.h"
#include "vl/Mat3.h"
#include<time.h>
#include<iostream>
#include<fstream>
#include<vector>
#include<list>

using namespace std;

class Triangle {
public:
  Triangle(Vec3f a,Vec3f b,Vec3f c,Vec3f rhoVal,Vec3f emmVal)
    :v1(a),v2(b),v3(c),rhoVal(rhoVal),emmVal(emmVal) {
    normal=norm(cross(b-a,c-a));
  }
  Vec3f v1;
  Vec3f v2;
  Vec3f v3; 
  Vec3f normal;
  Vec3f rhoVal;
  Vec3f emmVal;
  float area(){
#define a v1
#define b v2
#define c v3
    return len(cross(a,b)+cross(b,c)+cross(c,a))/2;
#undef a
#undef b
#undef c
}
};

typedef list<Triangle*> TriList;
typedef list<Triangle*>::iterator TriListItr;

long int subdivideUniform(float);
long int subdivideNonUniform(float);
void create_Point_Model();

TriList triangles;

int main(int argc, char* argv[])
{
  long int count=0;
  float maxarea = atof(argv[2]);
  int y = atoi(argv[3]);
  ifstream in(argv[1]);
  while(in){
    Vec3f a,b,c;
    Vec3f rhoVal;
    Vec3f emmVal;
    for(int j=0;j<3;j++)
      in>>a[j];
    for(int j=0;j<3;j++)
      in>>b[j];
    for(int j=0;j<3;j++)
      in>>c[j];
    for(int j=0;j<3;j++)
      in>>rhoVal[j];
    for(int j=0;j<3;j++)
      in>>emmVal[j];
    triangles.push_back(new Triangle(a,b,c,rhoVal,emmVal));
  }
 cout << "Read the triangles \n";

in.close();
if(y==1)
	count = subdivideUniform(maxarea);
else if(y==0)
	count = subdivideNonUniform(maxarea);
cout << " Number of triangles are : " <<count<<"\n";
create_Point_Model();
} 



long int subdivideUniform(float maxArea){
  long int cnt=0;
  TriListItr itr=triangles.begin();
  while(itr!=triangles.end()){
    if ((*itr)->area()<maxArea)
      itr++;
    else{
      Vec3f a=(*itr)->v1;
      Vec3f b=(*itr)->v2;
      Vec3f c=(*itr)->v3;
      Vec3f d=(a+b)/2.0;
      Vec3f e=(b+c)/2.0;
      Vec3f f=(c+a)/2.0;
      triangles.push_back(new Triangle(a,d,f,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(d,b,e,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(f,e,c,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(d,e,f,(*itr)->rhoVal,(*itr)->emmVal));
      itr=triangles.erase(itr);
      
    }
  }
  return(triangles.size());
}

long int subdivideNonUniform(float maxArea){
  long int cnt=0;
  TriListItr itr=triangles.begin();
  while(itr!=triangles.end()){
    if ((*itr)->area()<maxArea)
      itr++;
    else{
      Vec3f a=(*itr)->v1;
      Vec3f b=(*itr)->v2;
      Vec3f c=(*itr)->v3;
      Vec3f d,e,f;
	float val = rand()/((double)RAND_MAX+1);
      if(val>0.5) {
	val = val - .5;
	d=((1+val) * a+b)/(2+val);
	}
	else {
	d = (a+(1+val)*b)/(2+val);
	}
      val = rand()/((double)RAND_MAX+1);
      if(val>0.5) {
        val = val - .5;
        e=((1+val) * b+c)/(2+val);
        }
        else {
        e = (b+(1+val)*c)/(2+val);
        }
      val = rand()/((double)RAND_MAX+1);
      if(val>0.5) {
        val = val - .5;
        f=((1+val) * c+a)/(2+val);
        }
        else {
        f = (c+(1+val)*a)/(2+val);
        }
      triangles.push_back(new Triangle(a,d,f,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(d,b,e,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(f,e,c,(*itr)->rhoVal,(*itr)->emmVal));
      triangles.push_back(new Triangle(d,e,f,(*itr)->rhoVal,(*itr)->emmVal));
      itr=triangles.erase(itr);

    }
  }
  return(triangles.size());
}


void create_Point_Model()
{
 cout<<"Now in create_Point_Model() function \n";
 FILE *fp = fopen("pointModel_raf","w");
 for(TriListItr itr=triangles.begin();itr!=triangles.end();itr++)
 {
      Vec3f a=(*itr)->v1;
      Vec3f b=(*itr)->v2;
      Vec3f c=(*itr)->v3;
      Vec3f d=(a*2 + b + c)/3.0;
      Vec3f e=(a + 2*b + c)/3.0;
      Vec3f f=(a + b + 2*c)/3.0;      

 fprintf(fp,"%2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f \n",d[0],d[1],d[2],(*itr)->normal[0],(*itr)->normal[1],(*itr)->normal[2],(*itr)->rhoVal[0],(*itr)->rhoVal[1],(*itr)->rhoVal[2],(*itr)->emmVal[0],(*itr)->emmVal[1],(*itr)->emmVal[2]);
 fprintf(fp,"%2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f \n",e[0],e[1],e[2],(*itr)->normal[0],(*itr)->normal[1],(*itr)->normal[2],(*itr)->rhoVal[0],(*itr)->rhoVal[1],(*itr)->rhoVal[2],(*itr)->emmVal[0],(*itr)->emmVal[1],(*itr)->emmVal[2]);
  fprintf(fp,"%2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f %2f \n",f[0],f[1],f[2],(*itr)->normal[0],(*itr)->normal[1],(*itr)->normal[2],(*itr)->rhoVal[0],(*itr)->rhoVal[1],(*itr)->rhoVal[2],(*itr)->emmVal[0],(*itr)->emmVal[1],(*itr)->emmVal[2]);
}

fclose(fp);
}






